#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
GPT2 SeeKeR Agent for Language Modeling.
"""
from typing import Optional, Type, Tuple, List, Dict, Any
import torch
import torch.nn

from parlai.agents.fid.fid import FidModel, concat_enc_outs
from parlai.agents.hugging_face.gpt2 import GPT2Decoder, Gpt2Agent
from parlai.agents.rag.args import RetrieverType
from parlai.agents.rag.modules import RagModel
from parlai.agents.rag.retrievers import (
    Document,
    BLANK_DOC,
    SearchQuerySearchEngineRetriever,
    retriever_factory,
    RagRetriever,
)
from parlai.core.dict import DictionaryAgent
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from projects.seeker.agents.seeker_modules import (
    ComboFidSearchQuerySearchEngineRetriever,
)
from projects.seeker.agents.seeker_modules import (
    combo_fid_retriever_factory,
    interleave_fid_combo_outputs,
    ComboFidModel,
)

########################
# Retrieval Components #
########################


class FilterDocsForLabelSearchEngineRetrieverMixin:
    """
    Mixin for filtering the document containing the label out from retrieved docs.

    As these models are generally used with a CC knowledge dump, and the task on which
    they are evaluated is from the same CC dump, it is useful to filter out the document
    containing the label string from the retrieved documents, for a fair comparison.
    """

    def set_labels(self, labels: torch.Tensor):
        """
        Cache the label vec.
        """
        self.label_vec = labels

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Override retrieve and score to filter out docs that contain the label string.

        Copy over the whole thing because we need to check before chunking.
        """
        # step 1
        search_queries = self.generate_search_query(query)  # type: ignore

        # step 2
        search_results_batch = self.search_client.retrieve(
            search_queries, self.n_docs
        )  # type: ignore

        # step 3
        top_docs = []
        top_doc_scores = []
        max_n_docs: int = self.n_docs  # type: ignore
        for batch_id, (sq, search_results) in enumerate(
            zip(search_queries, search_results_batch)
        ):
            if not search_results:
                search_results = self._empty_docs(self.n_docs)  # type: ignore
            elif len(search_results) < self.n_docs:  # type: ignore
                remain_docs = self.n_docs - len(search_results)  # type: ignore
                search_results.extend(self._empty_docs(remain_docs))  # type: ignore
            docs_i = []
            scors_i = []
            # Change this debug later
            logging.debug(
                f'URLS:\n{self._display_urls(search_results)}'
            )  # type: ignore
            label_text = self.dict.vec2txt(
                self.label_vec[batch_id, :-1]
            )  # type: ignore
            for i, doc in enumerate(search_results):
                url = doc['url']
                title = doc['title']
                dcontent = doc['content']
                assert type(dcontent) in (
                    str,
                    list,
                ), f'Unrecognized retrieved doc: {dcontent}'
                full_text = (
                    dcontent if isinstance(dcontent, str) else '\n'.join(doc['content'])
                )
                if label_text in full_text:
                    docs_i.append(BLANK_DOC)
                    scors_i.append(0)
                else:
                    doc_chunks = [
                        dc[0]
                        for dc in self.pick_chunk(
                            sq, title, full_text, url
                        )  # type: ignore
                    ]
                    for splt_id, splt_content in enumerate(doc_chunks):
                        docs_i.append(
                            Document(
                                docid=url, text=splt_content, title=f'{title}_{splt_id}'
                            )
                        )
                        scors_i.append(self.rank_score(i))  # type: ignore
            max_n_docs = max(max_n_docs, len(docs_i))
            top_docs.append(docs_i)
            top_doc_scores.append(scors_i)
        # Pad with empty docs
        for i in range(len(top_docs)):
            n_empty = max_n_docs - len(top_docs[i])
            if n_empty:
                top_docs[i] = top_docs[i] + [BLANK_DOC] * n_empty
                top_doc_scores[i] = top_doc_scores[i] + [0] * n_empty
        self.top_docs = top_docs
        self.search_queries = search_queries
        return top_docs, torch.Tensor(top_doc_scores).to(query.device)


class FilterDocsForLabelSearchEngineRetrieverCombo(
    FilterDocsForLabelSearchEngineRetrieverMixin,
    ComboFidSearchQuerySearchEngineRetriever,
):
    pass


class FilterDocsForLabelSearchEngineRetrieverBase(
    FilterDocsForLabelSearchEngineRetrieverMixin, SearchQuerySearchEngineRetriever
):
    pass


#######################################
# Special Modules/Agents for GPT2 FiD #
#######################################


class IdentityLayer(torch.nn.Module):
    """
    Acts as the encoder for the FiD model.

    Custom Identity layer (as opposed to parlai.utils.torch.IdentityLayer) to account
    for special output required.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, *args, **kwargs):
        super().__init__()
        self.pad_idx = dictionary[dictionary.null_token]

    def forward(self, input: torch.LongTensor, *args, **kwargs):
        return input, input.ne(self.pad_idx)


class ComboGPT2Decoder(GPT2Decoder):
    """
    Allows us to use standard FiD/RAG model building logic.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args[:2])


#############################
# GPT2 With Retriever Model #
#############################


class GPT2WithRetrieverModel(FidModel):
    """
    GPT2 with Retriever Model.

    A GPT2 model that receives documents from a retriever.
    """

    def __init__(self, opt, dictionary, retriever_shared=None):
        self.add_start_token = opt["add_start_token"]
        FidModel.__init__(self, opt, dictionary, retriever_shared)
        self.config = self.seq2seq_decoder.transformer.config
        self.embedding_size = self.config.n_embd
        self.lm_head = torch.nn.Linear(
            self.config.n_embd, self.config.vocab_size, bias=False
        )
        self._tie_weights(self.lm_head, self.seq2seq_decoder.transformer.wte)
        self.doc_delim = self.dict.txt2vec(Document.PASSAGE_DELIM)[0]
        self.min_doc_len = opt['min_doc_token_length']
        self.truncate = (
            opt['text_truncate'] if opt['text_truncate'] > -1 else opt['truncate']
        )

    @classmethod
    def build_retriever(
        cls,
        opt: Opt,
        dictionary: DictionaryAgent,
        retriever_shared: Optional[Dict[str, Any]],
    ) -> Optional[RagRetriever]:
        if opt.get('filter_docs_with_label'):
            assert (
                RetrieverType(opt['rag_retriever_type']) == RetrieverType.SEARCH_ENGINE
            )
            return FilterDocsForLabelSearchEngineRetrieverBase(
                opt, dictionary, shared=retriever_shared
            )  # type: ignore
        else:
            return retriever_factory(opt, dictionary, shared=retriever_shared)

    @classmethod
    def build_encoder(
        cls,
        opt: Opt,
        *args,
        dictionary: Optional[DictionaryAgent] = None,
        embedding: Optional[torch.nn.Embedding] = None,
        encoder_class: Optional[Type] = None,
        **kwargs,
    ):
        """
        Override to build with IdentityLayer as Encoder.
        """
        return FidModel.build_encoder(
            opt, dictionary, encoder_class=IdentityLayer, **kwargs
        )

    @classmethod
    def build_decoder(
        cls,
        opt: Opt,
        *args,
        embedding: Optional[torch.nn.Embedding] = None,
        n_positions: Optional[int] = None,
        decoder_class: Optional[Type] = None,
        **kwargs,
    ):
        """
        Override to use GPT2 as decoder.
        """
        return FidModel.build_decoder(
            opt,
            Gpt2Agent.dictionary_class()(opt),
            decoder_class=ComboGPT2Decoder,
            **kwargs,
        )

    def encoder(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        target_lengths: Optional[torch.LongTensor],
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        Optional[List[List[Document]]],
        Optional[torch.Tensor],
    ]:
        """
        Override FidModel.encoder to pack all the documents into one input example.

        :param input:
            2D [bsz, seqlen] input to the encoder
        :param input_lengths:
            1D [bsz] lengths of each input item
        :param query_vec:
            2D [bsz*n_turns, seqlen] input for the retriever
        :param input_turns_cnt:
            1D [bsz] number of dialogue turns for each input example
        :param input_lengths:
            1D [bsz] lengths of each target item (for each input item)

        :return (encoder_out, encoder_mask, input_turns_cnt, top_docs, top_doc_scores):
            encoder_out: *concatenated* encoded representations of context/document pairs
            encoder_mask: new mask for enc_out
            input_turns_cnt: pass along the input turns count for the decoder
            top_docs: List of top Documents for each batch example
            top_doc_scores: scores for each retrieved document.
        """
        enc_out, mask, input_turns_cnt, top_docs, top_doc_scores = RagModel.encoder(
            self, input, input_lengths, query_vec, input_turns_cnt, positions, segments
        )  # type: ignore
        seq_len, n_docs = enc_out.size(1), enc_out.size(0) // input.size(0)

        if input_turns_cnt is not None:
            # Input Turns is a tensor of dim [bsz]
            input = input.repeat_interleave(input_turns_cnt, dim=0)  # type: ignore
        doc_starts = (enc_out == self.pad_idx).sum(dim=1)  # left padded
        doc_lens = (seq_len - input_lengths.repeat_interleave(n_docs)) - doc_starts
        # if no padding, docs are assumed to be min doc length long
        doc_lens[doc_lens.le(0)] = self.min_doc_len
        new_enc_out = enc_out.clone()
        # BEFORE:
        #  [
        #   pad...doc_0 / in_0
        #   pad...doc_1 / in_0
        #   ...
        #   pad...doc_n / in_m
        #                       ]
        total_length_i = 0
        for i, doc_len in enumerate(doc_lens):
            if i % n_docs == 0:
                total_length_i = 0
            # max doc length is determined by how much space we have after subtracting input length
            input_and_target = input_lengths[i // n_docs] + (
                target_lengths[i // n_docs] if target_lengths is not None else 0
            )
            max_doc_len = torch.div(
                (seq_len - input_and_target), n_docs, rounding_mode='floor'
            )
            max_doc_len = max(max_doc_len, self.min_doc_len)  # type: ignore
            doc_len = min(doc_len, max_doc_len)
            total_length_i += doc_len
            if i % n_docs == n_docs - 1:
                # keep the actual input context when processing the last doc.
                clamped_input_length = input_lengths[i // n_docs].clamp(
                    max=self.truncate - total_length_i
                )
                if target_lengths is not None:
                    clamped_input_length = input_lengths[i // n_docs].clamp(
                        max=self.truncate - total_length_i - target_lengths
                    )

                pad_end = seq_len - clamped_input_length - doc_len
            else:
                pad_end = seq_len - doc_len
            # we simply move the portion of the doc we want to keep to the end of the tensor,
            # and mask out everything else.
            mask[i, :pad_end] = False
            new_enc_out[i, pad_end : pad_end + doc_len] = enc_out[
                i, doc_starts[i] : doc_starts[i] + doc_len
            ]

        new_out, new_mask = concat_enc_outs(
            input, new_enc_out.unsqueeze(-1), mask, 1, self.pad_idx, right_padded=False
        )
        # After:
        #  [
        #   doc_0 / doc_1 / doc_2 ... in_0
        #   doc_0 / doc_1 / doc_2 ... in_m
        #                                   ]
        return new_out, new_mask, input_turns_cnt, top_docs, top_doc_scores

    def decoder(
        self,
        input: torch.LongTensor,
        encoder_state: Tuple[Any, ...],
        incr_state: Optional[Dict[str, Any]] = None,
    ) -> Tuple[torch.Tensor, Optional[Dict[str, Any]]]:
        """
        Override to make sure dimensions of enc_out are appropriate.

        :param input:
            input for the decoder
        :param encoder_state:
            RAG encoder states
        :param incr_state:
            incremental decoder state

        :return (output, new_incr_state):
            return the output token distribution, as well as new incremental state.
        """
        enc_out, *_ = encoder_state
        if enc_out.dim() == 2:
            enc_out = enc_out.t()
        else:
            enc_out = enc_out.squeeze(-1)
        dec_out, incr_state = self.seq2seq_decoder(
            input, enc_out, incr_state
        )  # type: ignore
        return dec_out, incr_state

    def concat_docs_and_input(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        top_docs: List[List[Document]],
        max_num_docs: int,
        right_padded: bool = True,
    ) -> torch.LongTensor:
        """
        Override to account for left-padded input.
        """
        return super().concat_docs_and_input(
            input, input_lengths, top_docs, max_num_docs, right_padded=False
        )

    def set_labels(self, labels: torch.LongTensor):
        """
        Set retriever's labels.
        """
        assert self.retriever is not None
        self.retriever.set_labels(labels)

    ### HFGPT2Model Functions

    def _get_decoder(self, opt, dict):
        return GPT2Decoder(opt, dict)

    def _tie_weights(self, output_embeddings, input_embeddings):
        output_embeddings.weight = input_embeddings.weight

    def _get_special_tokens(self, opt, dict):
        return dict.null_idx, dict.start_idx, dict.end_idx

    def output(self, tensor):
        """
        Compute output logits.
        """
        return self.lm_head(tensor)

    def reorder_decoder_incremental_state(self, incremental_state, inds):
        new_incr_state = []
        for layer_past in incremental_state:
            if torch.is_tensor(layer_past):
                new_incr_state.append(torch.index_select(layer_past, 1, inds))
            else:
                # newer versions of HF split up the intermediate outputs
                assert isinstance(layer_past, tuple)
                layer_past = torch.stack(layer_past, dim=0)
                new_incr_state.append(torch.index_select(layer_past, 1, inds))

        return tuple(new_incr_state)

    def decode_forced(self, encoder_states, ys):
        """
        Override to get rid of start token input.
        """
        if self.add_start_token:
            return super().decode_forced(encoder_states, ys)
        seqlen = ys.size(1)
        inputs = ys.narrow(1, 0, seqlen - 1)
        latent, _ = self.decoder(inputs, encoder_states)
        logits = self.output(latent)
        _, preds = logits.max(dim=2)
        return logits, preds


####################
# Combo GPT2 Model #
####################


class ComboGPT2Model(GPT2WithRetrieverModel, ComboFidModel):
    """
    Combo model.

    This model can handle contexts with and without retrieved documents.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, retriever_shared=None):
        super().__init__(opt, dictionary, retriever_shared)
        if opt.get('filter_docs_with_label'):
            assert (
                RetrieverType(opt['rag_retriever_type']) == RetrieverType.SEARCH_ENGINE
            )
            self.retriever = FilterDocsForLabelSearchEngineRetrieverCombo(
                opt, dictionary, shared=retriever_shared
            )  # type: ignore
        else:
            self.retriever = combo_fid_retriever_factory(
                opt, dictionary, shared=retriever_shared
            )
        self.top_docs: List[List[Document]] = []

    def get_top_docs(self) -> List[List[Document]]:
        """
        Return Documents.
        """
        return self.top_docs

    def encoder(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        skip_retrieval_vec: torch.BoolTensor,
        target_lengths: torch.LongTensor,
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        Optional[List[List[Document]]],
        Optional[torch.Tensor],
    ]:
        """
        Copy most of the ComboFidModel.forward here, but include the target_lengths for
        the encoder.

        Note, however, that we don't use it.
        """
        # If we have a full batch of retrieve (or not), can use simple logic here.
        if torch.all(skip_retrieval_vec):
            new_out, new_mask = self.seq2seq_encoder(input, positions, segments)
            return (
                new_out.unsqueeze(-1),
                new_mask,
                None,
                None,
                None,
            )  # need to unsqueeze to handle in decoder
        elif torch.all(~skip_retrieval_vec):
            output = super().encoder(
                input,
                input_lengths,
                query_vec,
                input_turns_cnt,
                None,
                positions,
                segments,
            )
            self.top_docs = output[-2]  # type: ignore
            return output
        assert all(t is None for t in [input_turns_cnt, positions, segments])
        # Encode with `super()` call for non-skip-retrieval inputs
        (
            enc_out_retrieval,
            mask_retrieval,
            _,
            top_docs,
            top_doc_scores,
        ) = super().encoder(
            input[~skip_retrieval_vec],  # type: ignore
            input_lengths[~skip_retrieval_vec],  # type: ignore
            query_vec[~skip_retrieval_vec],  # type: ignore
            input_turns_cnt,
            None,  # don't pass target lengths when combining
            positions,
            segments,
        )
        # Encode with seq2seq_encoder for skip-retrieval inputs
        enc_out_skip_retrieval, mask_skip_retrieval = self.seq2seq_encoder(
            input[skip_retrieval_vec]
        )
        (
            new_out,
            new_mask,
            new_top_docs,
            new_top_doc_scores,
        ) = interleave_fid_combo_outputs(
            enc_out_retrieval,
            enc_out_skip_retrieval.unsqueeze(-1),
            mask_retrieval,
            mask_skip_retrieval,
            skip_retrieval_vec,
            top_docs,  # type: ignore
            top_doc_scores,  # type: ignore
            right_padded=False,
        )
        self.top_docs = new_top_docs
        return new_out, new_mask, input_turns_cnt, new_top_docs, new_top_doc_scores



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from copy import deepcopy
from typing import Any, Dict, List, Set, Union, Optional, Tuple
import json
import logging
import os

from parlai.core.message import Message
from parlai.core.metrics import F1Metric
from parlai.core.params import ParlaiParser, Opt
from parlai.core.teachers import DialogTeacher
from parlai.utils.data import DatatypeHelper
import parlai.tasks.jericho_world.constants as consts

from .build import build, DATASET_NAME


def get_dtype(opt):
    return DatatypeHelper.fold(opt.get('datatype', 'train'))


def _path(opt):
    build(opt)
    dpath = os.path.join(opt['datapath'], DATASET_NAME)
    dtype = get_dtype(opt)
    if dtype == 'valid':
        logging.warning(
            'This data set does not have valid split. Using `test` instead.'
        )
        dtype = 'test'
    return os.path.join(dpath, f'{dtype}.json')


def clean_text(text: str) -> str:
    """
    Removes extra spaces and new lines from the text.
    """
    return text.replace('\n', ' ').strip()


def wrap_content(content: str, content_type: str) -> str:
    """
    Wraps content in tokens that shows its beginning and the end.
    """
    s = f'__{content_type}__'
    e = f'__end-{content_type}__'
    return f'{s} {content} {e}'


def graph_edge_as_str(subj: str, rel: str, obj: str) -> str:
    """
    Generate a formatted string from edge tuple components.
    """
    return '< ' + f' {consts.GRAPH_DELIM} '.join([subj, rel, obj]) + ' >'


def knowledge_graph_as_str(graph):
    if not graph:
        return consts.EMPTY_GRAPH_TOKEN

    graph_comps = []
    for edge_info in graph:
        # NOTE: often there are incomplete nodes that we detect and skip.
        # (Github issue reported here: https://github.com/JerichoWorld/JerichoWorld/issues/3)
        processed_edge = [e.strip() for e in edge_info if e.strip()]
        if len(processed_edge) == 3:
            s, r, o = processed_edge
            graph_comps.append(graph_edge_as_str(s, r, o))
    return (
        consts.SET_MEMBERS_DELIM.join(graph_comps)
        if graph_comps
        else consts.EMPTY_GRAPH_TOKEN
    )


def break_knowledge_graph(graph_str: str) -> Set[str]:
    return set([e.strip() for e in graph_str.split(consts.SET_MEMBERS_DELIM)])


def graph_mutation_diff(source_graph, dest_graph):
    """
    Generates the set of operations (ADD, DEL) needed to go from `source_graph` to
    `dest_graph`.
    """
    source_graph_edges = break_knowledge_graph(source_graph)
    dest_graph_edges = break_knowledge_graph(dest_graph)
    diff_edges = source_graph_edges.symmetric_difference(dest_graph_edges)

    mutations = []
    for edge in diff_edges:
        if edge in source_graph_edges:
            op = consts.GraphMutations.DEL
        else:
            op = consts.GraphMutations.ADD
        mutations.append(f'{op.name} {edge}')

    return mutations


def encode_set_elements(set1: Set[str], set2: Set[str]) -> Tuple[List[str]]:
    """
    Encodes (maps to int indices) elements in the union of two sets.
    """
    set1_enc = []
    set2_enc = []
    for el_id, el in enumerate(set1.union(set2)):
        el_id_str = str(el_id)
        if el in set1:
            set1_enc.append(el_id_str)
        if el in set2:
            set2_enc.append(el_id_str)
    return set1_enc, set2_enc


def extract_state_data(example_state: Dict, delim: str = ' ') -> Dict:
    def concat_vals(d):
        return [delim.join(p) for p in d.values()]

    return {
        'location_name': example_state['location']['name'],
        'observation': clean_text(example_state['obs']),
        'location_desc': clean_text(example_state['loc_desc']),
        'surrounding_objs': concat_vals(example_state['surrounding_objs']),
        'inventory_objs': concat_vals(example_state['inv_objs']),
        'valid_acts': list(example_state['valid_acts'].values()),
        'graph': example_state['graph'],
    }


class BaseJerichoWorldTeacher(DialogTeacher):
    """
    The base class that loads the games and episodes of the JerichoWorld.

    Note: do not use this class directly.
    """

    def __init__(self, opt: Opt, shared=None):
        opt = deepcopy(opt)
        opt['datafile'] = _path(opt)
        self.datatype = get_dtype(opt)
        self._incld_loc_name = opt['include_location']
        self.delim = opt['delimiter']
        self._incld_teacher_token = opt['include_teacher_token']
        self._incld_loc_desc = opt['include_location_description']
        self._incld_surr_objs = opt['include_surrounding_objects']
        self._prune_kg = opt['prune_knowledge_graph']
        self._keep_inv_during_kg_prune = False
        self.keep_next_state = True
        super().__init__(opt, shared=shared)

    def get_id(self):
        return 'JerichoWorldtBase'

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        arg_group = parser.add_argument_group('Jericho World Args')
        arg_group.add_argument(
            '--include-location',
            type='bool',
            default=True,
            help='Whether to include the name of the location',
        )
        arg_group.add_argument(
            '--include-teacher-token',
            type='bool',
            default=True,
            help='Whether to include the name of the teacher',
        )
        arg_group.add_argument(
            '--include-location-description',
            type='bool',
            default=True,
            help='Whether to include the text description of the location',
        )
        arg_group.add_argument(
            '--include-surrounding-objects',
            type='bool',
            default=False,
            help='Whether to include the list of surrounding objects',
        )
        arg_group.add_argument(
            '--prune-knowledge-graph',
            type='bool',
            default=False,
            help='If true, items in knowledge graph that were not in the state description will be eliminiated.',
        )
        arg_group.add_argument(
            '--delimiter',
            type=str,
            default='\n',
            help='Delimiter string to use between features',
        )

    def _clean_example(self, game_step: Dict[str, Any]) -> Dict[str, Any]:
        """
        Keeps data from example that we need, and discars the rest.
        """
        example = {
            'action': game_step['action'],
            'state': extract_state_data(game_step['state']),
        }
        if self.keep_next_state:
            example['next_state'] = extract_state_data(game_step['next_state'])
        return example

    def extract_knowledge_graph_str(
        self, state: Dict[str, Any], include_inv_objs: bool = False
    ) -> str:
        """
        Generates the string representing knowledge graph.

        It may modifiy the knowledge graph, for example when --prune-knowledge-graph is
        true.
        """

        def has_word_overlap(main_text, content_text_tokens):
            if not main_text or not content_text_tokens:
                return False

            for word in main_text.lower().split():
                if not word or len(word) < 3 or word in consts.GRAPH_VERT_SKIP_TOKEN:
                    continue
                if word in content_text_tokens:
                    return True
            return False

        def keep_edge(edge, text_tokens):
            """
            Returns false for edges that can NOT be readily inferred from the context.

            Edges that are referring to objects not referenced in the location
            description, except player (you). Thus, we drop these edges. Also, for some
            teachers (eg, ActionKGTeacher) the model can see user inventory before, and
            after the action, and they are important in result of action,. Thus, we keep
            < you , have, X > edges as well.
            """
            # Each graph edge is a tuple: (subject, relation, object)
            sub, rel, obj = [s.strip().lower() for s in edge]

            if sub == 'you':
                # player location
                if rel == 'in':
                    return True
                # user inventory object
                elif rel == 'have' and self._keep_inv_during_kg_prune:
                    return True

            return has_word_overlap(sub, text_tokens) and has_word_overlap(
                obj, text_tokens
            )

        if self._prune_kg:
            # Prunning the knowledge graph for the entities that are mentioned in the description
            old_graph = state['graph']
            new_graph = []

            # `text_desc` is the context that we show the model the generate the knowledge graph from.
            text_desc = f'{state["location_desc"]} {state["location_name"]}'
            if include_inv_objs:
                text_desc = f'{text_desc} {state["inventory_objs"]}'
            text_desc_tokens = text_desc.lower().split()

            # We prune the knowledge graph for items that are partially mentioned in `text_desc`
            for edge in old_graph:
                if keep_edge(edge, text_desc_tokens):
                    new_graph.append(edge)

            state['graph'] = new_graph

        return knowledge_graph_as_str(state['graph'])

    def generate_example_text_parts(self, example: Union[Dict, Message]) -> List[str]:
        """
        Returns the expected parts of the text for your example.

        These parts will be joined with `delim` to generate the text string.
        """
        teacher_token = (
            wrap_content(self.get_id(), 'tt') if self._incld_teacher_token else ''
        )

        curr_state = example['state']
        return [
            teacher_token,
            self.location_context(curr_state),
            self.surrounding_objects_context(curr_state),
        ]

    @abc.abstractmethod
    def generate_example_label(self, example: Union[Dict, Message]) -> str:
        """
        Implement this method for the expected label of your example.
        """

    def generate_example_label_candidates(
        self, example: Union[Dict, Message]
    ) -> List[str]:
        """
        Override this to include a list of label candidates with the example.
        """
        return []

    def location_context(self, example_state: Dict) -> str:
        """
        Generates the context text for the location.
        """
        loc_context = []

        if self._incld_loc_name:
            loc_context.append(
                wrap_content(example_state['location_name'], consts.LOCATION_NAME)
            )

        if self._incld_loc_desc:
            loc_context.append(
                wrap_content(
                    clean_text(example_state['location_desc']),
                    consts.LOCATION_DESCRIPTION,
                )
            )
        return ' '.join(loc_context)

    def surrounding_objects_context(self, example_state: Dict) -> str:
        out = ''
        if self._incld_surr_objs:
            content_str = consts.SET_MEMBERS_DELIM.join(
                example_state['surrounding_objs']
            )
            out = wrap_content(content_str, consts.SURROUNDING_OBJECTS)
        return out

    def knowledge_graph_context(self, example_state: Dict) -> str:
        """
        Generate the context for the flattened knowledge graph.
        """
        knowledge_graph = self.extract_knowledge_graph_str(example_state)
        return wrap_content(knowledge_graph, consts.KNOWLEDGE_GRAPH)

    def load_data(self, datafile):
        logging.info(f'Reading data from {datafile}')
        with open(datafile) as df:
            return json.load(df)

    def skip_example(self, example: Dict) -> bool:
        """
        Implements checks that some teachers have on valid examples, based on content.
        """
        return False

    def check_fix_state_kg(self, game: Dict, game_step_id: int) -> None:
        """
        Tries to fix the KG for the state associate with game_step_id, in place.

        It replaces the KG for game_step_id with the KG for `next_step` from previous
        step, or uses the `graph_diff` if game_step_id == 0.
        """
        if game[game_step_id]['state']['graph']:
            return

        if game_step_id == 0:
            # The first empty step, replace with the diff.
            game[0]['state']['graph'] = game[0]['graph_diff']
        else:
            # Get the graph from the next_state that was created in the previous step.
            prev_state_next = game[game_step_id - 1]['next_state']
            game[game_step_id]['state']['graph'] = deepcopy(prev_state_next['graph'])

    def _generate_example_text(self, text_parts: List[str]) -> str:
        """
        Concatnates the non-empty parts of text fetrue into a single string.
        """
        return self.delim.join([s.strip() for s in text_parts if s.strip()])

    def setup_data(self, datafile: str):
        for game in self.load_data(datafile):
            for step_i, game_step in enumerate(game):
                self.check_fix_state_kg(game, step_i)
                example = self._clean_example(game_step)
                if self.skip_example(example):
                    continue
                example['text'] = self._generate_example_text(
                    self.generate_example_text_parts(example)
                )

                # Preparing the label (received as str or list)
                label = self.generate_example_label(example)
                if not isinstance(label, list):
                    assert isinstance(label, str)
                    label = [label]
                example['labels'] = label

                label_candts = self.generate_example_label_candidates(example)
                if label_candts:
                    example['label_candidates'] = label_candts
                    if label not in label_candts:
                        # Making sure label is present in the candidates. Often the exact label
                        # may not be in the list of actions.
                        # This might be because of the short version or rephrase of it being used.
                        label_candts.extend(label)

                new_episode = step_i == 0
                yield example, new_episode


class BaseJerichoWorldSingleEpisodeTeacher(BaseJerichoWorldTeacher):
    """
    Truns each examples into a single episode: No history.
    """

    def setup_data(self, datafile: str):
        for example, _ in super().setup_data(datafile):
            yield example, True


###############################################################
#                                                             #
# Knowledge Graph and state prdiction teachers                #
#                                                             #
###############################################################


class StateToKGTeacher(BaseJerichoWorldSingleEpisodeTeacher):
    """
    Game state to the knowledge graph.
    """

    def get_id(self):
        return 'StateKG'

    def generate_example_label(self, example: Union[Dict, Message]) -> str:
        return self.extract_knowledge_graph_str(example['state'])

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ) -> None:
        if model_response.is_padding() or (not model_response.get('text', None)):
            return

        expected_graph = break_knowledge_graph(labels[0].lower())
        predicted_graph = break_knowledge_graph(model_response['text'].lower())

        # Encoding the graph edges/mutation operations into ints for readily use of F1Metric
        expected_graph_enc, predicted_graph_enc = encode_set_elements(
            expected_graph, predicted_graph
        )
        self.metrics.add(
            'response_elements_f1',
            F1Metric.compute(
                guess=' '.join(predicted_graph_enc),
                answers=[' '.join(expected_graph_enc)],
            ),
        )

        # Subject, Relation F1
        # Changind "(MUT) < you , in , house >"   --into-->   "(MUT) < you , in "
        # This is to check F1 for the predicted subject and relation overlap.
        ekg_sub_rel = set([e.rsplit(',', 1)[0] for e in expected_graph])
        pkg_sub_rel = set([e.rsplit(',', 1)[0] for e in predicted_graph])
        ekg_sub_rel_ids, pkg_sub_rel_ids = encode_set_elements(ekg_sub_rel, pkg_sub_rel)
        self.metrics.add(
            'graph_subject_relation_f1',
            F1Metric.compute(
                guess=' '.join(pkg_sub_rel_ids), answers=[' '.join(ekg_sub_rel_ids)]
            ),
        )

        # Subject F1
        # Changind "(MUT) < you , in " (produced above)   --into-->   "(MUT) < you "
        # This is to check F1 for the predicted subject overlap.
        ekg_sub = set([e.split(',')[0] for e in ekg_sub_rel])
        pkg_sub = set([e.split(',')[0] for e in pkg_sub_rel])
        ekg_sub_ids, pkg_sub_ids = encode_set_elements(ekg_sub, pkg_sub)
        self.metrics.add(
            'graph_subject_f1',
            F1Metric.compute(
                guess=' '.join(pkg_sub_ids), answers=[' '.join(ekg_sub_ids)]
            ),
        )


class StaticKGTeacher(StateToKGTeacher):
    """
    Generates the knowledge graph from a single state.
    """

    def get_id(self):
        return 'StaticStateKG'

    def skip_example(self, example: Dict) -> bool:
        return (
            self.extract_knowledge_graph_str(example['state'])
            == consts.EMPTY_GRAPH_TOKEN
        )


class ActionKGTeacher(StateToKGTeacher):
    """
    Generates the knowledge graph mutations after a given action.
    """

    def __init__(self, opt: Opt, shared=None):
        self._keep_inv_during_kg_prune = True
        super().__init__(opt, shared=shared)

    def get_id(self):
        return 'Action2KGMutation'

    def skip_example(self, example: str):
        for st in ('state', 'next_state'):
            if (
                self.extract_knowledge_graph_str(example[st])
                == consts.EMPTY_GRAPH_TOKEN
            ):
                return True
        return False

    def generate_example_text_parts(self, example: Union[Dict, Message]) -> List[str]:
        prts = super().generate_example_text_parts(example)

        # Adding knowledge graph to the text.
        prts.append(self.knowledge_graph_context(example['state']))

        # Adding observation after the action to the text.
        prts.append(
            wrap_content(example['next_state']['observation'], consts.OBSERVATION)
        )

        prts.append(wrap_content(example['action'], consts.ACTION))
        return prts

    def generate_example_label(self, example: Union[Dict, Message]) -> str:
        curr_graph = self.extract_knowledge_graph_str(
            example['state'], include_inv_objs=True
        )
        next_graph = self.extract_knowledge_graph_str(
            example['next_state'], include_inv_objs=True
        )
        graph_diff = graph_mutation_diff(curr_graph, next_graph)
        return (
            # sorting to pass the tests, otherwise the results are in various orders.
            '\n'.join(sorted(graph_diff))
            if graph_diff
            else consts.GraphMutations.NO_MUTATION.name
        )


###############################################################
#                                                             #
# Action prediction teachers                                  #
#                                                             #
###############################################################


class ActionPredictionBase(BaseJerichoWorldSingleEpisodeTeacher):
    """
    The base class for predicting the actions (or set of valid action).

    This is an abstract class, use its children for predicting actions.
    """

    def __init__(self, opt: Opt, shared=None):
        self._incld_knwldg_grph = opt['include_knowledge_graph']
        self._keep_inv_during_kg_prune = True
        super().__init__(opt, shared=shared)

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt=None) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        arg_group = parser.add_argument_group('Jericho World Action Teacher Args')
        arg_group.add_argument(
            '--include-knowledge-graph',
            type='bool',
            default=True,
            help='Whether to include the knowledge graph.',
        )

    def generate_example_text_parts(self, example: Union[Dict, Message]) -> List[str]:
        prts = super().generate_example_text_parts(example)
        if self._incld_knwldg_grph:
            prts.append(self.knowledge_graph_context(example['state']))
        return prts


class StateToValidActionsTeacher(ActionPredictionBase):
    """
    Generates the set of valid actions.
    """

    def get_id(self):
        return 'ValidActionsTeacher'

    def generate_example_label(self, example: Union[Dict, Message]) -> List[str]:
        return example['state']['valid_acts']


class StateToActionTeacher(ActionPredictionBase):
    """
    Game state to action (selecting between the set of valid actions).
    """

    def get_id(self):
        return 'NextActionTeacher'

    def generate_example_label_candidates(
        self, example: Union[Dict, Message]
    ) -> List[str]:
        """
        Override this to include a list of label candidates with the example.
        """
        return example['state']['valid_acts']

    def generate_example_label(self, example: Union[Dict, Message]) -> str:
        return example['action']


class DefaultTeacher(StaticKGTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import nltk
import os
import random
import copy
from typing import List, Optional, Tuple

from parlai.core.agents import create_agent_from_model_file, Agent
from parlai.core.build_data import modelzoo_path
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
from parlai.core.metrics import F1Metric
from parlai.core.mutators import register_mutator, MessageMutator, ManyEpisodeMutator
from parlai.core.opt import Opt
import parlai.core.tod.tod_core as tod
from parlai.tasks.blended_skill_talk.build import build as build_bst
import parlai.tasks.wizard_of_internet.constants as CONST

from parlai.utils.strings import normalize_reply

from projects.seeker.utils import calc_f1_msc, IS_SEARCH_REQUIRED
from projects.seeker.tasks.mutators import (
    PromptSearchQueryMutator,
    ExtractEntityResponse,
)

import projects.bb3.constants as BB3_CONST
import projects.bb3.prompts as PROMPT


###################
# Prompt Mutators #
###################


@register_mutator('prompt_knowledge_mutator')
class PromptKnowledgeMutator(PromptSearchQueryMutator):
    """
    Add a __knowledge__ prompt to the end of the context, to inform the model.

    Assumes flattened data.
    """

    PROMPT = BB3_CONST.GENERATE_KNOWLEDGE


@register_mutator('prompt_memory_mutator')
class PromptMemoryMutator(PromptSearchQueryMutator):
    """
    Add a __gnerate-memory__ prompt to the end of the context, to inform the model.

    Assumes flattened data.
    """

    PROMPT = BB3_CONST.GENERATE_MEMORY


@register_mutator('prompt_extract_entity_mutator')
class PromptEntityExtractionMutator(PromptSearchQueryMutator):
    """
    Add a __extract-entity__ prompt to the end of the context, to inform the model.

    Assumes flattened data.
    """

    PROMPT = BB3_CONST.EXTRACT_ENTITY


@register_mutator('prompt_memory_decision_mutator')
class PromptMemoryDecisionMutator(PromptSearchQueryMutator):
    """
    Add a __is-memory-required__ prompt to the end of the context, to inform the model.

    Assumes flattened data.
    """

    PROMPT = BB3_CONST.IS_MEMORY_REQUIRED


#############################
# Memory Knowledge Mutators #
#############################


SUMMARIZER = None


ALL_PERSONAS = None


def build_summarizer(opt: Opt) -> Agent:
    """
    Build the Persona Summarizer.
    """
    return create_agent_from_model_file(
        modelzoo_path(opt['datapath'], 'zoo:bb3/persona_summarizer/model'),
        opt_overrides={
            'skip_generation': False,
            'inference': 'beam',
            'beam_size': 3,
            'beam_block_ngram': 3,
            'beam_min_length': 10,
            'datatype': 'valid',
        },
    )


def build_all_personas(opt: Opt) -> List[str]:
    """
    Build the personas list from which we sample memories for MDM.
    """
    personas_path = os.path.join(
        opt['datapath'], 'blended_skill_talk', 'persona_list.txt'
    )
    if not os.path.exists(personas_path):
        new_opt = copy.deepcopy(opt)
        new_opt['task'] = 'blended_skill_talk'
        build_bst(new_opt)
    with open(personas_path) as f:
        all_personas = [l.replace('||', '').replace('\n', '') for l in f.readlines()][
            :-1
        ]
    return all_personas


def merge_personas(personas: List[str], p_to_merge: List[List[str]]) -> List[str]:
    """
    Merge two groups of personas, based on word overlap.

    :param personas:
        list of original personas.
    :param p_to_merge:
        list of personas to merge.
        first element is partner personas, second is your personas.

    :return merged:
        return list of merged personas
    """
    new_personas = []
    split_personas = [
        [p.replace('\n', ' ') for p in personas if p.startswith('partner')],
        [p.replace('\n', ' ') for p in personas if p.startswith('your')],
    ]
    for i in range(2):
        prefix = "partner's persona: " if i == 0 else "your persona: "
        personas_i = split_personas[i]
        for ip in p_to_merge[i]:
            found = False
            for j, op in enumerate(personas_i):
                if F1Metric.compute(ip, [op[op.index(':') + 2 :]]) > 0.5:
                    personas_i[j] = ' '.join([op, ip])
                    found = True
                    break
            if not found:
                personas_i.append(f"{prefix}{ip}")
        new_personas.append(personas_i)
    return new_personas[0] + new_personas[1]


def get_overlap_sentence(
    full_text: str, label: str, docs: List[str], find_speaker: bool = True
) -> Tuple[str, int, Optional[int]]:
    """
    Get the sentence that most overlaps with a sentence in the context.

    :param full_text:
        full context
    :param label:
        target to find overlap with
    :param docs:
        list of candidate strings for overlap
    :param find_speaker:
        whether to determine who said the sentence in the full text

    :return (best_sentence, best_f1, best_idx):
        if we reach the F1 threshold, return the
        corresponding sentence, F1 score, and index into the docs
    """
    best_f1 = 0
    best_sentence = ''
    best_idx = None
    try:
        gold_parts = nltk.word_tokenize(label)
    except IndexError:
        return best_sentence, best_f1, best_idx
    for i, d in enumerate(docs):
        ds = d.split('\n')
        for s in ds:
            f1 = calc_f1_msc(s, gold_parts)
            if f1 > best_f1:
                best_f1 = f1
                best_sentence = s
                best_idx = i
    if best_sentence != '' and find_speaker:
        # find which speaker it is
        z = full_text.split('\n')
        z.reverse()
        ind = z.index(best_sentence)
        if (ind % 2) == 0:
            speaker = '__them__'
        else:
            speaker = '__you__'
        if 'your persona:' in best_sentence:
            speaker = '__you__'

        best_sentence = f"{best_sentence} {speaker}"

    return best_sentence, best_f1, best_idx


@register_mutator("personas_as_docs")
class PersonasAsDocsMutator(ManyEpisodeMutator):
    """
    This mutator does the following:

    1. Computes memories for context lines with the persona summarizer
    2. Determines which memory has highest overlap with target label
    3. Depending on target, does the following:
        - if knowledge, we set the target as that persona memory
        - if memory_decision, set the target as access_memory if enough overlap, otherwise don't
    """

    THRESHOLD = 0.3
    TARGET = 'knowledge'

    def __init__(self, opt: Opt) -> None:
        super().__init__(opt)
        assert 'flatten' not in opt['mutators'], 'cannot use flatten with this mutator.'
        global SUMMARIZER
        if SUMMARIZER is None:
            SUMMARIZER = build_summarizer(opt)
        global ALL_PERSONAS
        if ALL_PERSONAS is None:
            ALL_PERSONAS = build_all_personas(opt)

    def compute_context_memories(
        self, episode: List[Message]
    ) -> Tuple[List[str], List[str]]:
        """
        Compute the memories from the context.

        :param episode:
            list of messages

        :return (self_memories, partner_memories):
            return list of memories for self and partner.
        """
        partner_texts = [m['text'] for m in episode]
        first_text = [
            t for t in partner_texts[0].split('\n') if not t.startswith('your persona:')
        ]
        if not first_text:
            partner_texts[0] = BB3_CONST.DUMMY_TEXT
        else:
            partner_texts[0] = first_text[0]
        self_texts = [m.get('labels', m.get('eval_labels'))[0] for m in episode]

        assert SUMMARIZER is not None
        partner_memories = SUMMARIZER.batch_respond(
            [
                Message({'text': p.split('\n')[-1], 'episode_done': True})
                for p in partner_texts
            ]
        )
        self_memories = SUMMARIZER.batch_respond(
            [Message({'text': s, 'episode_done': True}) for s in self_texts]
        )

        return self_memories, partner_memories

    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        assert ALL_PERSONAS is not None
        global NUM_DONE

        def _maybe_add_persona(text: str, is_partner: bool):
            prefix = "partner's persona: " if is_partner else "your persona: "
            if text != BB3_CONST.NOPERSONA:
                raw_personas.append(text)
                all_personas.append(f"{prefix}{text}")

        # compute the personas
        self_memories, partner_memories = self.compute_context_memories(episode)

        # find best overlaps
        all_personas = []
        raw_personas = []
        new_episode = []
        all_text = []
        for i, msg in enumerate(episode):
            text = msg['text']
            label = msg['labels'][0]
            if i == 0:
                if 'persona:' in text:
                    # Convai2; session 1
                    personas = [
                        t for t in text.split('\n') if t.startswith('your persona:')
                    ]
                    text = [
                        t for t in text.split('\n') if not t.startswith('your persona:')
                    ]
                    if not text:
                        text = BB3_CONST.DUMMY_TEXT
                        msg.force_set('text', text)
                    else:
                        text = text[0]
                    all_personas += personas
                    raw_personas += [
                        p.replace('your persona: ', '').replace(
                            "partner's persona: ", ''
                        )
                        for p in all_personas
                    ]
                elif 'personas' in msg:
                    # MSC
                    all_personas += msg['personas'].split('\n')
                    init_personas = msg['init_personas']
                    all_personas = merge_personas(all_personas, init_personas)
                    raw_personas = [
                        p.replace('your persona: ', '').replace(
                            "partner's persona: ", ''
                        )
                        for p in all_personas
                    ]
            assert (
                'persona:' not in text
            ), "cannot use this mutator with teacher that puts personas in context"
            all_text.append(text)
            best_sentence, best_f1, best_idx = get_overlap_sentence(
                '\n'.join(all_text), msg['labels'][0], raw_personas, False
            )
            if best_f1 > self.THRESHOLD:
                assert best_idx is not None
                if self.TARGET == 'knowledge':
                    msg.force_set(CONST.RETRIEVED_DOCS, all_personas)
                    msg.force_set(CONST.RETRIEVED_DOCS_URLS, [''] * len(all_personas))
                    msg.force_set(CONST.RETRIEVED_DOCS_TITLES, [''] * len(all_personas))
                    msg.force_set(CONST.SELECTED_SENTENCES, [all_personas[best_idx]])
                    msg.force_set('raw_personas', raw_personas)
                    msg.force_set('old_target', msg['labels'])
                    msg.force_set('labels', [best_sentence])
                    msg.force_set('text', '\n'.join(all_text))
                elif self.TARGET == 'memory_decision':
                    msg.pop('personas', None)
                    msg.pop('init_personas', None)
                    msg.force_set('persona_sentence', best_sentence)
                    msg.force_set('old_target', msg['labels'])
                    msg.force_set('labels', [BB3_CONST.DO_ACCESS_MEMORY])
                    # make sure to include memory in the context!
                    if not best_sentence.startswith('persona:'):
                        best_sentence = f"persona: {best_sentence}"
                    msg.force_set(
                        'text', "\n".join([best_sentence, all_text[-1].split('\n')[-1]])
                    )
                new_episode.append([msg])
            elif self.TARGET == 'memory_decision':
                msg.pop('personas', None)
                msg.pop('init_personas', None)
                msg.force_set('persona_sentence', best_sentence)
                msg.force_set('old_target', msg['labels'])
                msg.force_set('labels', [BB3_CONST.DONT_ACCESS_MEMORY])
                # include a random memory in context!
                random_pers = (
                    random.choice(ALL_PERSONAS)
                    .replace('your persona:', 'persona:')
                    .replace("partner's persona:", 'persona:')
                )
                if not random_pers.startswith('persona:'):
                    random_pers = f"persona: {random_pers}"
                msg.force_set(
                    'text', "\n".join([random_pers, all_text[-1].split('\n')[-1]])
                )
                new_episode.append([msg])
            _maybe_add_persona(partner_memories[i], True)
            _maybe_add_persona(self_memories[i], False)
            all_text.append(label)

        return new_episode


@register_mutator("convert_overlap_to_personas_as_docs")
class ConvertOverlapToPersonasAsDocs(ManyEpisodeMutator):
    """
    This mutator converts a knowledge task that originally computed word overlap between
    target and all prior sentences to choose a previous utterance as the knowledge
    sentence.

    The conversion turns all context utterances into persona memories (retrieved docs),
    and all targets into persona memories as well.
    """

    TARGET = 'knowledge'

    def __init__(self, opt: Opt) -> None:
        super().__init__(opt)
        global SUMMARIZER
        if SUMMARIZER is None:
            SUMMARIZER = build_summarizer(opt)

    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        assert len(episode) == 1
        message = episode[0]
        texts = message['text'].split('\n')
        texts[-1] = texts[-1].replace(f' {BB3_CONST.GENERATE_KNOWLEDGE}', '')
        all_personas = []
        if 'persona:' in message['text']:
            # Convai2; session 1
            personas = [t for t in texts if t.startswith('your persona:')]
            new_text = [t for t in texts if not t.startswith('your persona:')]
            if not new_text:
                new_text = BB3_CONST.DUMMY_TEXT
            message.force_set('text', '\n'.join(new_text))
            all_personas += personas
        elif 'personas' in message:
            # MSC
            all_personas += message['personas'].split('\n')
        assert SUMMARIZER is not None
        required_summaries = [
            Message(
                {
                    'text': t.replace(f' {BB3_CONST.GENERATE_KNOWLEDGE}', ''),
                    'episode_done': True,
                }
            )
            for t in texts
        ] + [
            Message(
                {
                    'text': message['labels'][0]
                    .replace(f' {BB3_CONST.YOU}', '')
                    .replace(f' {BB3_CONST.THEM}', '')
                }
            )
        ]
        summaries = []
        offset = 16
        for start_idx in range(0, len(required_summaries), offset):
            summaries += SUMMARIZER.batch_respond(
                required_summaries[start_idx : start_idx + offset]
            )
        self_summaries = [
            t
            for i, t in enumerate(reversed(summaries[:-1]))
            if i % 2 == 1 and BB3_CONST.NOPERSONA not in t
        ]
        partner_summaries = [
            t
            for i, t in enumerate(reversed(summaries[:-1]))
            if i % 2 == 0 and BB3_CONST.NOPERSONA not in t
        ]
        all_personas = merge_personas(all_personas, [partner_summaries, self_summaries])
        if summaries[-1] != BB3_CONST.NOPERSONA or 'persona:' in message['labels'][0]:
            # for Convai2 we might already have this.
            new_target = (
                message['labels'][0]
                if 'persona:' in message['labels'][0]
                else summaries[-1]
            )
            if 'persona:' not in new_target:
                prefix = (
                    "your" if BB3_CONST.YOU in message['labels'][0] else "partner's"
                )
                new_target = f"{prefix} persona: {new_target}"
            message.force_set('original_target', message['labels'])
            message.force_set(CONST.RETRIEVED_DOCS, all_personas)
            message.force_set(CONST.RETRIEVED_DOCS_URLS, [''] * len(all_personas))
            message.force_set(CONST.RETRIEVED_DOCS_TITLES, [''] * len(all_personas))
            message.force_set(CONST.SELECTED_SENTENCES, [new_target])
            message.force_set('labels', [new_target])
            message.force_set(
                'text',
                message['text'].replace(
                    BB3_CONST.GENERATE_KNOWLEDGE, BB3_CONST.ACCESS_MEMORY
                ),
            )
            message.pop('skip_retrieval', None)
            return [[message]]
        else:
            return []


@register_mutator('prompt_access_memory_mutator')
class PromptAccessMemoryMutator(PromptSearchQueryMutator):
    """
    Add a __access-memory__ prompt to the end of the context, to inform the model.

    Assumes flattened data.
    """

    PROMPT = BB3_CONST.ACCESS_MEMORY


@register_mutator("match_target_to_persona_mutator")
class MatchTargetToPersonaMutator(MessageMutator):
    """
    For ad-hoc conversion; matches the knowledge label to a "your" or "partner's"
    prefix.
    """

    def message_mutation(self, message: Message) -> Message:
        personas = message[CONST.RETRIEVED_DOCS]
        target = message['labels'][0]
        p = None
        for p in personas:
            if target in p:
                break
        assert p
        assert target in p
        message.force_set('labels', [p])
        return message


@register_mutator("fix_mkm_formatting_mutator")
class FixMkmFormattingMutator(MessageMutator):
    """
    Ad-hoc mutator for the utterance overlap Mkm teachers.

    With ConvAI2 and BST, the utterance overlap mutators found persona sentences from the context.

    But those examples were not formatted correctly with `convert_overlap_to_personas_as_docs`

    So, fixing that now
    """

    def message_mutation(self, message: Message) -> Message:
        text = message['text']
        if not text.endswith(BB3_CONST.ACCESS_MEMORY):
            message.force_set('text', f"{text} {BB3_CONST.ACCESS_MEMORY}")
        for tok in [BB3_CONST.YOU, BB3_CONST.THEM]:
            if tok in message['labels'][0]:
                message.force_set(
                    'labels', [message['labels'][0].replace(f' {tok}', '')]
                )
                assert tok not in message['labels'][0]
        return message


############################
# Memory Decision Mutators #
############################
@register_mutator("memory_decision_mutator")
class MemoryDecisionMutator(PersonasAsDocsMutator):
    """
    Utilize mutator above to set the memory decision.
    """

    TARGET = 'memory_decision'


############################
# Style Grounding Mutators #
############################
@register_mutator("style_gen_to_grm")
class StyleGenToVRMMutator(MessageMutator):
    """
    Converts style gen tasks to grounded dialogue tasks.

    assumes flattened data
    """

    def message_mutation(self, message: Message) -> Message:
        text = message['text']
        style = message['personality']
        assert BB3_CONST.BEGIN_STYLE not in text
        message.force_set(
            'text', f"{text}\n{BB3_CONST.BEGIN_STYLE} {style} {BB3_CONST.END_STYLE}"
        )
        return message


############################
# Search Dialogue Mutators #
############################
@register_mutator("funpedia_to_bb3_mutator")
class FunpediaToBB3WithStyleMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        context = message['text'].split('\n')
        assert len(context) == 3
        topic, style, knowledge = context
        message.force_set(
            'text',
            f"{topic}\n{BB3_CONST.TOKEN_KNOWLEDGE} {knowledge} {BB3_CONST.TOKEN_END_KNOWLEDGE}"
            f"\n{BB3_CONST.BEGIN_STYLE} {style} {BB3_CONST.END_STYLE}",
        )
        return message


@register_mutator("tod_to_srm_mutator")
class TodToSRMMutator(ManyEpisodeMutator):
    """
    Converts TOD Tasks to a Search Dialogue Response task.

    System grounding on the API responses.

    Flattens the episodes as well
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        context = []
        knowledge: List = []
        system_silence = f"{tod.STANDARD_SYSTEM_UTTERANCE}{BB3_CONST.DUMMY_TEXT}"
        user_silence = f"{tod.STANDARD_USER_UTTERANCE}{BB3_CONST.DUMMY_TEXT}"
        for m in episode:
            new_message = m.copy()
            user_utt: str = m['text']
            system_utt: str = m['labels'][0]
            system_utt_clean = system_utt.replace(
                tod.STANDARD_SYSTEM_UTTERANCE, ''
            ).replace('\n', ' ')
            if user_utt.startswith(tod.STANDARD_API_SCHEMAS):
                # start of convo; don't need this
                continue
            elif (
                user_utt.startswith(tod.STANDARD_USER_UTTERANCE)
                and user_utt != user_silence
            ):
                # add user utterances to the conversation
                context.append(user_utt.replace(tod.STANDARD_USER_UTTERANCE, ''))
            if user_utt.startswith(tod.STANDARD_RESP) and user_utt != tod.STANDARD_RESP:
                # here's the knowledge grounding
                knowledge.append(
                    user_utt.replace(tod.STANDARD_RESP, '').replace('\n', '')
                )

            if knowledge and not (
                system_utt == system_silence or system_utt.startswith(tod.STANDARD_CALL)
            ):
                str_knowledge = f"{BB3_CONST.TOKEN_KNOWLEDGE} {'. '.join(knowledge)} {BB3_CONST.TOKEN_END_KNOWLEDGE}"
                new_message.force_set('text', '\n'.join(context + [str_knowledge]))
                new_message.force_set('labels', [system_utt_clean])
                if F1Metric.compute(system_utt_clean, [str_knowledge]) > 0.0:
                    # filter out pathological examples
                    out_episodes.append([new_message])
                # reset knowledge
                knowledge = []
            if (
                not system_utt.startswith(tod.STANDARD_CALL)
                and system_utt != system_silence
            ):
                context.append(system_utt_clean)

        return out_episodes


############################
# Memory Dialogue Mutators #
############################
@register_mutator("convert_mkm_to_mrm_mutator")
class ConvertMkmToMrmMutator(MessageMutator):
    """
    Converts mkm task to mrm task.
    """

    def message_mutation(self, message: Message) -> Message:
        persona_knowledge = message['labels'][0]
        message.force_set(
            'text',
            f"{message['text'].replace(f' {BB3_CONST.ACCESS_MEMORY}', '')}\n{BB3_CONST.BEGIN_MEMORY} {persona_knowledge} {BB3_CONST.END_MEMORY}",
        )
        old_target = message['old_target']
        if not isinstance(old_target, list):
            old_target = [old_target]
        message.force_set('labels', old_target)
        return message


############################
# Entity Dailogue Mutators #
############################
@register_mutator("extract_entity_for_response_model_bb3")
class ExtractEntityResponseBB3(ExtractEntityResponse):
    BEGIN_KNOWLEDGE: str = BB3_CONST.BEGIN_ENTITY
    END_KNOWLEDGE: str = BB3_CONST.END_ENTITY


###################
# Filter Mutators #
###################


@register_mutator("filter_silence_only_mutator")
class FilterSilenceOnlyMutator(ManyEpisodeMutator):
    """
    Filters out episodes that only contain silence.
    """

    def many_episode_mutation(self, episode):
        for m in episode:
            text = m['text']
            for tok in [
                IS_SEARCH_REQUIRED,
                BB3_CONST.IS_MEMORY_REQUIRED,
                BB3_CONST.ACCESS_MEMORY,
            ]:
                text = text.replace(f' {tok}', '')
            if text.lower() == '__silence__':
                return []
        return [episode]


@register_mutator("filter_silence_only_memory_decision_mutator")
class FilterSilenceOnlyMemoryDecisionMutator(ManyEpisodeMutator):
    """
    Filters out episodes that only contain silence.
    """

    def many_episode_mutation(self, episode):
        for m in episode:
            text = m['text'].split('\n')
            if len(text) == 2 and text[1].lower() == '__silence__':
                return []
        return [episode]


@register_mutator("filter_wow_topic_only_search_decision_mutator")
class FilterWowTopicOnlySearchDecisionMutator(ManyEpisodeMutator):
    """
    Filters out episodes from the search decision teacher that are only the chosen
    topic.
    """

    def many_episode_mutation(self, episode):
        for m in episode:
            text = m['text'].replace(f' {IS_SEARCH_REQUIRED}', '')
            if 'chosen_topic' in m and text == m['chosen_topic']:
                return []
            elif len(text.split()) < 3:
                return []
        return [episode]


#######################
# Formatting Mutators #
#######################
@register_mutator("ensure_same_number_docs_and_titles_mutator")
class EnsureSameNumberDocsAndTitles(MessageMutator):
    """
    Ensures that docs and doc titles have same number of items.
    """

    def message_mutation(self, message: Message) -> Message:
        docs = message[CONST.RETRIEVED_DOCS]
        for key in [CONST.RETRIEVED_DOCS_URLS, CONST.RETRIEVED_DOCS_TITLES]:
            if len(message[key]) < len(docs):
                message.force_set(
                    key, message[key] + [''] * (len(docs) - len(message[key]))
                )
            elif len(message[key]) > len(docs):
                message.force_set(key, message[key][: len(docs)])
        return message


def _normalize_persona_line(x: str) -> str:
    """
    Normalize a persona line.
    """
    if x.startswith('your persona: '):
        # Normalize the sentence appearing after 'your persona:'
        x = x[len('your persona: ') :]
        x = normalize_reply(x)
        x = 'your persona: ' + x
    elif x.startswith("partner's persona: "):
        x = x[len("partner's persona: ") :]
        x = normalize_reply(x)
        x = "partner's persona: " + x
    return x


@register_mutator("normalize_reply_mutator")
class NormalizeReplyMutator(MessageMutator):
    """
    Uses string normalization over text and labels.

    And retrieved docs, I suppose...
    """

    def message_mutation(self, message: Message) -> Message:
        """
        Need to normalize:

        1) text 2) label 3) retrieved docs 4) best sentence.
        """
        # 1. text
        texts = message['text'].split('\n')
        your_personas = []
        partner_personas = []
        non_personas = []
        for i, x in enumerate(texts):
            if x.startswith('your persona: '):
                # Normalize the sentence appearing after 'your persona:'
                x = _normalize_persona_line(x)
                your_personas.append(x)
            elif x.startswith("partner's persona: "):
                x = _normalize_persona_line(x)
                partner_personas.append(x)
            elif i == len(texts) - 1:
                # check for memory decision, memory, generate memory etc.
                if BB3_CONST.IS_MEMORY_REQUIRED in x:
                    x = x.replace(f" {BB3_CONST.IS_MEMORY_REQUIRED}", '')
                    x = normalize_reply(x)
                    x = f"{x} {BB3_CONST.IS_MEMORY_REQUIRED}"
                    non_personas.append(x)
                elif BB3_CONST.ACCESS_MEMORY in x:
                    x = x.replace(f" {BB3_CONST.ACCESS_MEMORY}", '')
                    x = normalize_reply(x)
                    x = f"{x} {BB3_CONST.ACCESS_MEMORY}"
                    non_personas.append(x)
                elif BB3_CONST.BEGIN_MEMORY in x:
                    x = x.replace(f'{BB3_CONST.BEGIN_MEMORY} ', '').replace(
                        f' {BB3_CONST.END_MEMORY}', ''
                    )
                    if 'persona:' in x:
                        x = _normalize_persona_line(x)
                    else:
                        x = normalize_reply(x)
                    x = f"{BB3_CONST.BEGIN_MEMORY} {x} {BB3_CONST.END_MEMORY}"
                    non_personas.append(x)
                else:
                    x = normalize_reply(x)
                    non_personas.append(x)
            else:
                x = normalize_reply(x)
                non_personas.append(x)
        message.force_set(
            'text', '\n'.join(your_personas + partner_personas + non_personas)
        )
        # 2. label
        label = message['labels'][0]
        if 'persona:' in label:
            label = _normalize_persona_line(label)
        elif label not in [BB3_CONST.DONT_ACCESS_MEMORY, BB3_CONST.DO_ACCESS_MEMORY]:
            label = normalize_reply(label)
        message.force_set('labels', [label])

        # 3. retrieved docs.
        if CONST.RETRIEVED_DOCS in message:
            documents = message[CONST.RETRIEVED_DOCS]
            new_docs = []
            for x in documents:
                x = _normalize_persona_line(x)
                new_docs.append(x)
            message.force_set(CONST.RETRIEVED_DOCS, new_docs)
        # 4. best sentence
        if CONST.SELECTED_SENTENCES in message:
            selected = message[CONST.SELECTED_SENTENCES][0]
            message.force_set(
                CONST.SELECTED_SENTENCES, [_normalize_persona_line(selected)]
            )
        return message


@register_mutator("fits_pop_keys_mutator")
class FitsPopKeysMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        for key in ['bot_acts', 'bot_gold', 'human_acts']:
            message.pop(key, None)
        return message


@register_mutator("fits_remove_special_toks_mutator")
class FitsRemoveSpecialToksMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        from parlai.tasks.fits.mutators import GoldKnowledge

        message.force_set(
            'labels',
            [message['labels'][0].replace(f"{GoldKnowledge.TOKEN_DEC_KNOWLEDGE} ", '')],
        )
        message.force_set(
            'text', message['text'].replace(f" {GoldKnowledge.TOKEN_ENC_KNOWLEDGE}", '')
        )
        return message


########################
# OPT Teacher Mutators #
########################
@register_mutator("prefix_speakers")
class PrefixSpeakersMutator(MessageMutator):
    """
    Add control tokens to the speakers, within the context.
    """

    SELF_PREFIX = BB3_CONST.YOU_PREFIX
    PARTNER_PREFIX = BB3_CONST.PARTNER_PREFIX

    def message_mutation(self, message: Message) -> Message:
        context, utterances, post_utterances = self.get_context_and_utterances(message)
        context = [c.lstrip(' ').rstrip(' ') for c in context]
        utterances = [u.lstrip(' ').rstrip(' ') for u in utterances]
        post_utterances = [p.lstrip(' ').rstrip(' ') for p in post_utterances]

        new_utts: List[str] = []
        for i, utt in enumerate(reversed(utterances)):
            if utt.lower() != BB3_CONST.DUMMY_TEXT.lower() and not any(
                utt.startswith(tok)
                for tok in [
                    BB3_CONST.BEGIN_ENTITY,
                    BB3_CONST.BEGIN_MEMORY,
                    BB3_CONST.TOKEN_KNOWLEDGE,
                ]
            ):
                prefix = self.SELF_PREFIX if i % 2 == 1 else self.PARTNER_PREFIX
                utt = f"{prefix}{utt.lstrip(' ').rstrip(' ')}"
            new_utts.append(utt)
        for i in range(len(context)):
            # LIGHT context starts with underscores
            if 'persona:' not in context[i] and not context[i].startswith('_'):
                context[i] = f"{self.PARTNER_PREFIX}{context[i]}"
        message.force_set(
            'text', '\n'.join(context + list(reversed(new_utts)) + post_utterances)
        )
        return message

    def get_context_and_utterances(
        self, message: Message
    ) -> Tuple[List[str], List[str], List[str]]:
        context, utterances, post_utterances = [], [], []
        texts = message['text'].split('\n')
        m_id = message.get('id', '')
        # 1. Handle WizInt
        context_end = 0
        if ('Dialogue' in m_id and 'Vanilla' not in m_id) or (
            'msc_dialogue' in m_id
            and any(
                tok in message['text']
                for tok in [
                    BB3_CONST.BEGIN_MEMORY,
                    BB3_CONST.BEGIN_ENTITY,
                    BB3_CONST.TOKEN_KNOWLEDGE,
                ]
            )
        ):
            post_utterances = [texts[-1]]
            texts = texts[:-1]
        if 'Light' in m_id:
            context_end = 5
        elif 'Funpedia' in m_id:
            post_utterances.append(texts[-1])
            texts = texts[:-1]
        elif any(
            m_id.endswith(f'Woi{t}Teacher')
            for t in [
                'SearchDialogue',
                'VanillaDialogue',
                'SearchKnowledge',
                'SearchQuery',
            ]
        ):
            context_end = 2
        elif m_id.endswith('MemoryDecisionTeacher') or (
            'msc' in m_id and texts[0].startswith('persona:')
        ):
            context_end = 1
        elif any(
            m_id.endswith(f'Wow{t}Teacher')
            for t in ['SearchDialogue', 'VanillaDialogue', 'SearchKnowledge']
        ):
            context_end = 1
        elif any(
            m_id.endswith(f'BST{t}Teacher')
            for t in [
                'EntityDialogue',
                'VanillaDialogue',
                'StyleGroundingDialogue',
                'EntityKnowledge',
                'MemoryKnowledge',
                'MemoryKnowledgeUttOverlap',
            ]
        ):
            ctxt_dataset = message['context_dataset']
            if (
                any(tok in m_id for tok in ['MemoryKnowledge', 'MemoryDialogue'])
                and ctxt_dataset == 'wizard_of_wikipedia'
            ):
                context_end = 1
            elif any(
                t in m_id
                for t in [
                    'Knowledge',
                    'Dialogue',
                    'VanillaDialogue',
                    'StyleGroundingDialogue',
                ]
            ) and any('persona:' in t for t in texts):
                non_context = [i for i, t in enumerate(texts) if 'persona:' not in t]
                end_idx = 0
                if ctxt_dataset == 'wizard_of_wikipedia':
                    end_idx = 1
                context_end = (
                    non_context[end_idx] if len(non_context) > end_idx else len(texts)
                )
        elif any(
            m_id.endswith(f'Convai2{t}Teacher')
            for t in [
                'EntityDialogue',
                'EntityKnowledge',
                'VanillaDialogue',
                'StyleGroundingDialogue',
            ]
        ):
            non_context = [i for i, t in enumerate(texts) if 'persona:' not in t]
            context_end = non_context[0] if non_context else len(texts)
            if 'StyleGrounding' in m_id:
                # this has the topic as well
                context_end += 1

        context = texts[:context_end]
        utterances = texts[context_end:] if len(texts) > context_end else []
        return context, utterances, post_utterances


@register_mutator("prefix_speakers_opt")
class PrefixSpeakersOPTMutator(PrefixSpeakersMutator):
    SELF_PREFIX = f"{PROMPT.SELF_PREFIX}: "
    PARTNER_PREFIX = f"{PROMPT.PARTNER_PREFIX}: "

    def message_mutation(self, message: Message) -> Message:
        message = super().message_mutation(message)
        m_id = message.get('id', '')
        if (
            'dialogue' in m_id.lower()
            and not ('NQOpen' in m_id)
            and not (
                ('msc' in m_id and message['text'].startswith('persona:'))
                or ('msc' in m_id and BB3_CONST.EXTRACT_ENTITY in message['text'])
                or ('msc' in m_id and message['text'].endswith(BB3_CONST.ACCESS_MEMORY))
                or (
                    'msc' in m_id
                    and message['text'].endswith(BB3_CONST.IS_SEARCH_REQUIRED)
                )
                or (
                    'msc' in m_id
                    and message['text'].endswith(BB3_CONST.IS_MEMORY_REQUIRED)
                )
            )
        ):
            if message['labels'][0].startswith(PROMPT.SELF_PREFIX):
                message.force_set(
                    'labels',
                    [message['labels'][0].replace(f"{PROMPT.SELF_PREFIX}: ", '')],
                )
            if not message['text'][0].endswith(f"{PROMPT.SELF_PREFIX}: "):
                message.force_set('text', f"{message['text']}\n{PROMPT.SELF_PREFIX}: ")
        return message


@register_mutator("format_decision_tasks_for_decoder_only")
class FormatDecisionTasksForDecoderOnlyMutator(MessageMutator):
    """
    Replaces special tokens with proper prefixes.

    Before:
        (Convai2MemoryDecisionBalancedWithMemoryTeacher)
        persona: i actually like wearing suits and ties.
        Then you must be working. Or just taking some time off? __is-memory-required__
            __do-not-access-memory__

        - - - NEW EPISODE: WoiSearchDecisionTeacher - - -
        he would be so rich, I wish I had that much money! __is-search-required__
            __do-search__

    After:
        (Convai2DecoderOnlyMemoryDecisionBalancedWithMemoryTeacher)
        Personal Fact: i actually like wearing suits and ties.
        Person 1: Then you must be working. Or just taking some time off?
        Memory Decision:
            do not access memory

        - - - NEW EPISODE: WoiDecoderOnlySearchDecisionTeacher - - -
        Person 1: he would be so rich, I wish I had that much money!
        Search Decision:
            search
    """

    def message_mutation(self, message: Message) -> Message:
        context, label = message['text'], message['labels'][0]
        assert any(
            const in context
            for const in [BB3_CONST.IS_SEARCH_REQUIRED, BB3_CONST.IS_MEMORY_REQUIRED]
        ), message['id']
        context = context.replace(
            BB3_CONST.IS_SEARCH_REQUIRED, f"\n{PROMPT.SEARCH_DECISION}: "
        )
        context = context.replace(
            BB3_CONST.IS_MEMORY_REQUIRED, f"\n{PROMPT.MEMORY_DECISION}: "
        )
        context = context.replace('persona: ', f"{PROMPT.MEMORY_KNOWLEDGE_PREFIX}: ")
        assert '\n\n' not in context
        label = label.replace(BB3_CONST.DO_SEARCH, PROMPT.SEARCH)
        label = label.replace(BB3_CONST.DO_NOT_SEARCH, PROMPT.DO_NOT_SEARCH)
        label = label.replace(BB3_CONST.DO_ACCESS_MEMORY, PROMPT.ACCESS_MEMORY)
        label = label.replace(BB3_CONST.DONT_ACCESS_MEMORY, PROMPT.DO_NOT_ACCESS_MEMORY)
        message.force_set('text', context)
        message.force_set('labels', [label])
        return message


@register_mutator("format_gen_tasks_for_decoder_only")
class FormatGenTasksForDecoderOnlyMutator(MessageMutator):
    """
    For use with SGM, and MGM.

    Replaces the special tokens with proper prefixes.

    Before:
        (MSCMemoryGeneratorTeacher)
        Hi how are you doing today?
        Doing good. Not ready for the weekend to be over though.
        So true. Off to work?
        Yeah, I am an accountant. It kinda runs in the family. Ll
        My brother is employed at best buy. It does not run in the family
        Ll. That's a good place to shop. Any pets? I have a couple dogs. __generate-memory__
            I have a couple of dogs.

        ---------------
        (WoiSearchQueryTeacher)
        I live in Mesa, Arizona.
        I just bought my first home.
        a bungalow close to the city __generate-query__
            arizona

    After:
        (MSCDecoderOnlyMemoryGeneratorTeacher)
        Person 2: Hi how are you doing today?
        Person 1: Doing good. Not ready for the weekend to be over though.
        Person 2: So true. Off to work?
        Person 1: Yeah, I am an accountant. It kinda runs in the family. Ll
        Person 2: My brother is employed at best buy. It does not run in the family
        Person 1: Ll. That's a good place to shop. Any pets? I have a couple dogs.
        Memory:
            I have a couple of dogs.
        -------------------
        (WoiDecoderOnlySearchQueryTeacher)
        Person 1: I live in Mesa, Arizona.
        Person 1: I just bought my first home.
        Person 1: a bungalow close to the city
        Query:
            arizona
    """

    def message_mutation(self, message: Message) -> Message:
        context, _ = message['text'], message['labels'][0]
        assert any(
            const in context
            for const in [BB3_CONST.GENERATE_MEMORY, BB3_CONST.GENERATE_QUERY]
        )
        context = context.replace(
            BB3_CONST.GENERATE_MEMORY, f"\n{PROMPT.MEMORY_GEN_PREFIX}: "
        )
        context = context.replace(
            BB3_CONST.GENERATE_QUERY, f"\n{PROMPT.QUERY_GEN_PREFIX}: "
        )
        assert '\n\n' not in context
        message.force_set('text', context)
        return message


@register_mutator("format_knowledge_tasks_for_decoder_only")
class FormatKnowledgeForDecoderOnlyMutator(MessageMutator):
    """
    Format examples for Decoder-Only models.

    This mutator places all of the retrieved documents within the context.

    The mutator ensures that the context + label does not exceed 2048, the
    context length of OPT models.

    Before:
        - - - NEW EPISODE: Convai2MemoryKnowledgeTeacher - - -
        Hi! I work as a gourmet cook. __access-memory__
            your persona: I hate carrots.

        - - - NEW EPISODE: WoiKnowledgeTeacher - - -
        My favorite athlete is nadal.
        he will be the greatest tennis player __generate-knowledge__
            Looking for a new hobby? Or trying to decide which one will be the best use of your limited free time? We’ve got you covered.

        - - - NEW EPISODE: Convai2KnowledgeTeacher - - -
        your persona: I had a gig at local theater last night.
        your persona: I work as a stand up comedian.
        your persona: I come from a small town.
        your persona: My favorite drink is cuba libre.
        your persona: I did a few small roles in tv series.
        We all live in a yellow submarine, a yellow submarine. Morning!
        Hi! That is a great line for my next stand up.
        Lol. I am shy, anything to break the ice, and I am a beatles fan. __extract-entity__
            tv

    After:
        - - - NEW EPISODE: Convai2DecoderOnlyMemoryKnowledgeTeacher - - -
        Person 2's Persona: I'm very athletic.
        Person 2's Persona: I wear contacts.
        Person 2's Persona: I have brown hair.
        Person 2's Persona: I love bicycling.
        Person 2's Persona: I hate carrots.
        Person 1's Persona: I work as a gourmet cook. I am female.
        Person 2's Persona: I don't like carrots. I throw them away.
        Person 1's Persona: I can sing. I can sing pitch perfect.
        Person 2's Persona: I ride my bike to work. I cook.
        Person 1's Persona: I won a spelling bee. I am a student.
        Person 2's Persona: I have a good sense of humor. I can see through what you are trying to sell me.
        Person 1's Persona: I was published in the new yorker once.
        Person 2's Persona: Spelling is important to me. I am a teacher.
        Person 1's Persona: I can cook. I am not a chef.
        Person 2's Persona: My hair is brown. I am white. I have brown hair.
        Person 1's Persona: I have no hair. I am asian.
        Person 2's Persona: I like hairless asians. I like carrots.
        Person 1's Persona: I eat carrots like a horse. I love carrots.
        Person 1's Persona: I work as a gourmet cook. I have a perfect voice.
        Person 1: Hi! I work as a gourmet cook.
        Personal Fact:
            Person 2's Persona: I hate carrots.

        - - - NEW EPISODE: WoiDecoderOnlyKnowledgeTeacher - - -
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | The Ultimate List of Hobbies
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | feeling it. Eventually you�
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | lucrative.. These hobbies are rewarding
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | to do it for free by
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Teaching and Tutoring Music.
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Tutoring Children. 54.
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | you can meditate.. 66
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | a penny for your thoughts.
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Learning how to DIY. 96
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | wonderful paper art.. List of
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | pay for itself if you review
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Robot Making. 148. Pipe
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | if you have a facility near
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Pursuing one of these hobbies
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | are plenty of guided yoga courses
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | drop-shipping, running
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | friends away at weddings.. 200
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | but all age groups can enjoy
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | and birdhouses. 238.
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | start, the better you�
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | more flexible than adults, so
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Producing Electronic Music. If
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | awesome.. 290. Soap
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | games are fun alone or with
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | of 53 Hobbies for Your
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Camping. 342. Sur
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Expand your horizons and experience
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Fell Running. 389. Track
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | always wanted to forge your own
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | as you once were. Brewing
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Snowmobiling. 422.
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | sculpture racing. 454.
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | season, you get delicious homegrown
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | or just looking to expand your
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | getting started. From there,
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Looking for a new hobby?
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | views and quiet time to yourself
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | you want to be outside,
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | comes from is worth it in
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | Maybe one of the most peaceful
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | lakes, ponds; anywhere there
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | a track. Find somewhere beautiful
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | not the prettiest thing,
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | gets to know the world around
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | up, but I guarantee that
        External Knowledge: The Ultimate List of Hobbies 549 handpicked hobby ideas | for history lovers. Find something
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Stamp Collecting: Also for anyone who
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | ways of life. Knowing how people dressed
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | intent. Even better, learn a language
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | from the next, so collect and press
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | and collect them. They are great decorations
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | games. It goes beyond the simple sport
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | and style: you may just learn a
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | better way to feel classy than by serving
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Bread Making: A great way to customize
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | like it, without being beholden to
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | world around you. Like writing, painting
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | time you’ll soon be making
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | design. A scrapbook allows to you
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | A practical form of art, quil
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | If you enjoy adorning yourself with
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | One of the more intensive yet possibly rewarding
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | if you really have a lot of time
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | for maximum fun.. 68. Furniture
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Build cabinets, fix plumbing, demolish
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Dawn. Get drawn into a fantastic interactive
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | true intellectual’s game the world
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | your friends? Juggling really does help
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | help you learn about the world and its
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | safe, but if you do it right
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | for a Historical Society: Love history and
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | have. Check out these guides to the
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | can’t get your voice on
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | entertainment.. 94. Dancing: Both art
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Nemo think so. In reality,
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | most of us have a lot to learn
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | you can dress and role play as them
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | want to prepare for Russian strategic rocket forces
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | /New Arrivals/. Taxi Car w
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Trailer Clear Display Case. 1/25
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Mega Hobby. /Our Family Business/.
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | at the top of our pyramid. But
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | A Truly Unimaginable List of
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | the other. They also bring different people
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | as a hobby, and is now a
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | and a true observer is the one that
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | who is known to have seen the maximum
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | that every human seeks because it gives them
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | College Dictionary.. Hobbies Related to Comput
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | of Facebook, started writing software as a
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | there are many people, who define relaxing
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | Fishing All Across Europe: Spain Has the
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | of the house, there are some who
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | card (received). Did you know?
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | for celebrities and models, or people who
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | how the 54-year-old div
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | very dear Tom Hanks, who collects
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | The world s premier supplier of radio remote
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | 16A 380W Power Supply. Red
        External Knowledge: 100 Creative Fun Hobbies to Try Ultimate List 2021 | electric aircraft.. Parts Saver ONLY $
        Person 1: My favorite athlete is nadal.
        Person 1: he will be the greatest tennis player
        Interesting Fact:
            Looking for a new hobby? Or trying to decide which one will be the best use of your limited free time? We’ve got you covered.

        - - - NEW EPISODE: Convai2DecoderOnlyKnowledgeTeacher - - -
        Person 2's Persona: I had a gig at local theater last night.
        Person 2's Persona: I work as a stand up comedian.
        Person 2's Persona: I come from a small town.
        Person 2's Persona: My favorite drink is cuba libre.
        Person 2's Persona: I did a few small roles in tv series.
        Person 1: We all live in a yellow submarine, a yellow submarine. Morning!
        Person 2: Hi! That is a great line for my next stand up.
        Person 1: Lol. I am shy, anything to break the ice, and I am a beatles fan.
        Previous Topic:
            tv
    """

    MAX_MODEL_LEN = 2048
    MAX_CONTEXT_LEN = 1024

    def __init__(self, opt):
        self.opt = opt
        self.dict = DictionaryAgent(
            {**ParlaiParser(True, True).parse_args([]), 'dict_tokenizer': 'gpt2'}
        )
        self.task = opt['task']

    def join_docs(self, docs: List[str], module: str) -> str:
        if module == 'search':
            prefix = PROMPT.EXTERNAL_KNOWLEDGE_PREFIX
            return f"{prefix}: " + f'\n{prefix}: '.join(docs)
        else:
            docs = [
                d.replace('your persona', PROMPT.SELF_MEMORY_PREFIX).replace(
                    'partner\'s persona', PROMPT.PARTNER_MEMORY_PREFIX
                )
                for d in docs
            ]
            return '\n'.join(docs)

    def reduce_doc_len(
        self,
        gen_prompt: str,
        context: str,
        joined_docs: str,
        message: Message,
        docs: List[str],
        module: str,
    ) -> str:
        prompt_toks_len = len(self.dict.txt2vec(gen_prompt))
        context_len = len(self.dict.txt2vec(context))
        docs_len = len(self.dict.txt2vec(joined_docs))
        label_len = len(self.dict.txt2vec(message['labels'][0]))
        max_doc_len = self.MAX_MODEL_LEN - context_len - label_len - prompt_toks_len
        if docs_len > max_doc_len:
            per_doc_len = max_doc_len // len(docs)
            docs = [self.dict.vec2txt(self.dict.txt2vec(d)[:per_doc_len]) for d in docs]
            joined_docs = self.join_docs(docs, module)
        return joined_docs

    def message_mutation(self, message: Message) -> Message:
        docs = message.get(CONST.RETRIEVED_DOCS)
        if (
            not docs
            or 'Entity' in message['id']
            or BB3_CONST.EXTRACT_ENTITY in message['text']
        ):
            # contextual knowledge
            text = message['text'].replace(
                f" {BB3_CONST.EXTRACT_ENTITY}",
                f"\n{PROMPT.CONTEXTUAL_KNOWLEDGE_PREFIX}: ",
            )
            # convert persona lines
            text = text.replace(
                "your persona:", f"{PROMPT.SELF_MEMORY_PREFIX}:"
            ).replace("partner's persona:", f"{PROMPT.PARTNER_MEMORY_PREFIX}")
            text = '\n'.join(
                [
                    c
                    for c in text.split('\n')
                    if BB3_CONST.DUMMY_TEXT.lower() not in c.lower()
                ]
            )
            message.force_set('text', text)
            return message
        context = (
            message['text']
            .replace(BB3_CONST.ACCESS_MEMORY, '')
            .replace(BB3_CONST.GENERATE_KNOWLEDGE, '')
            .replace(BB3_CONST.EXTRACT_ENTITY, '')
        )
        context = '\n'.join(
            [
                c
                for c in context.split('\n')
                if BB3_CONST.DUMMY_TEXT.lower() not in c.lower()
            ]
        )
        titles = message[CONST.RETRIEVED_DOCS_TITLES]
        if all(t for t in titles):
            docs = [f"{t} | {d}" for t, d in zip(titles, docs)]
            docs = [d.replace('\n', '. ') for d in docs]

        if any('persona:' in d for d in docs):
            module = 'memory'
        else:
            module = 'search'
        joined_docs = self.join_docs(docs, module)

        if module == 'search':
            gen_prompt = f"{PROMPT.SEARCH_KNOWLEDGE_PREFIX}: "
        else:
            gen_prompt = f"{PROMPT.MEMORY_KNOWLEDGE_PREFIX}: "
            label = (
                message['labels'][0]
                .replace('your persona', PROMPT.SELF_MEMORY_PREFIX)
                .replace('partner\'s persona', PROMPT.PARTNER_MEMORY_PREFIX)
            )
            message.force_set('labels', [label])

        joined_docs = self.reduce_doc_len(
            gen_prompt, context, joined_docs, message, docs, module
        )

        message.force_set('text', '\n'.join([joined_docs, context, gen_prompt]))

        return message


@register_mutator("format_knowledge_tasks_for_decoder_only_reduced_docs")
class FormatKnowledgeForDecoderOnlyReducedDocsMutator(
    FormatKnowledgeForDecoderOnlyMutator
):
    """
    Same as format_knowledge_tasks_for_decoder_only, except reduces the number of docs
    before reducing the doc length.
    """

    NUM_DOCS = 5
    CHUNK_SIZE = 500

    def message_mutation(self, message: Message) -> Message:
        from parlai.tasks.wizard_of_internet.mutators import chunk_docs_in_message

        message.force_set(
            CONST.SELECTED_SENTENCES,
            message.get(CONST.SELECTED_SENTENCES) or message['labels'],
        )
        message = chunk_docs_in_message(message, self.CHUNK_SIZE)
        docs = message.get(CONST.RETRIEVED_DOCS)
        checked_sentences = message.get(CONST.SELECTED_SENTENCES) or message['labels']
        if isinstance(checked_sentences, tuple):
            checked_sentences = [l for l in checked_sentences]
        for i in range(len(checked_sentences)):
            checked_sentences[i] = checked_sentences[i].lstrip(' ').rstrip(' ')
        assert docs
        assert checked_sentences
        context = (
            message['text']
            .replace(BB3_CONST.ACCESS_MEMORY, '')
            .replace(BB3_CONST.GENERATE_KNOWLEDGE, '')
            .replace(BB3_CONST.EXTRACT_ENTITY, '')
        )
        good_doc_indexes = [
            i for i, d in enumerate(docs) if any(c in d for c in checked_sentences)
        ]
        rest = [i for i in range(len(docs)) if i not in good_doc_indexes]
        if len(good_doc_indexes) < self.NUM_DOCS:
            remaining_left = self.NUM_DOCS - len(good_doc_indexes)
            if len(rest) >= remaining_left:
                good_doc_indexes += random.sample(rest, remaining_left)
            else:
                good_doc_indexes += rest

        docs = [d for i, d in enumerate(docs) if i in good_doc_indexes]
        titles = [
            t
            for i, t in enumerate(message[CONST.RETRIEVED_DOCS_TITLES])
            if i in good_doc_indexes
        ]
        if all(t for t in titles):
            docs = [f"{t} | {d}" for t, d in zip(titles, docs)]
            docs = [d.replace('\n', '. ') for d in docs]

        joined_docs = self.join_docs(docs, 'search')

        gen_prompt = f"{PROMPT.SEARCH_KNOWLEDGE_PREFIX}: "

        joined_docs = self.reduce_doc_len(
            gen_prompt, context, joined_docs, message, docs, 'search'
        )

        message.force_set('text', '\n'.join([joined_docs, context, gen_prompt]))

        return message


@register_mutator("format_response_tasks_for_decoder_only")
class FormatResponseTasksForDecoderOnlyMutator(MessageMutator):
    """
    Replace special tokens with appropriate prefixes.

    Before:
        - - - NEW EPISODE: WoiDialogueTeacher - - -
        My favorite game is WWE.
        fight game very interesting to played
        __knowledge__ Play WWE Games __endknowledge__
            That's really interesting! Are there are a lot WWE games out there? Why do you like them so much>

        - - - NEW EPISODE: Convai2DialogueTeacher - - -
        your persona: I had a gig at local theater last night.
        your persona: I work as a stand up comedian.
        your persona: I come from a small town.
        your persona: My favorite drink is cuba libre.
        your persona: I did a few small roles in tv series.
        We all live in a yellow submarine, a yellow submarine. Morning!
        Hi! That is a great line for my next stand up.
        Lol. I am shy, anything to break the ice, and I am a beatles fan.
        __entity__ tv __endentity__
            I can tell. I am not, you can see me in some tv shows

        - - - NEW EPISODE: Convai2DialogueFromPersonaOverlapMAMTeacher - - -
        Hi! I work as a gourmet cook.
        __memory__ your persona: I hate carrots. __endmemory__
            I don't like carrots. I throw them away.

    After:
        - - - NEW EPISODE: WoiDecoderOnlyDialogueTeacher - - -
        Person 1: My favorite game is WWE.
        Person 1: fight game very interesting to played
        Interesting Fact: Play WWE Games
            Person 2: That's really interesting! Are there are a lot WWE games out there? Why do you like them so much>

        - - - NEW EPISODE: Convai2DecoderOnlyDialogueTeacher - - -
        Person 1: We all live in a yellow submarine, a yellow submarine. Morning!
        Person 2: Hi! That is a great line for my next stand up.
        Person 1: Lol. I am shy, anything to break the ice, and I am a beatles fan.
        Previous Topic: tv
            Person 2: I can tell. I am not, you can see me in some tv shows

        - - - NEW EPISODE: Convai2DecoderOnlyDialogueFromPersonaOverlapMAMTeacher - - -
        Person 1: Hi! I work as a gourmet cook.
        Personal Fact: Person 2's Persona: I hate carrots.
            Person 2: I don't like carrots. I throw them away.
    """

    def message_mutation(self, message: Message) -> Message:
        context, _ = message['text'], message['labels'][0]
        assert any(
            const in context
            for const in [
                BB3_CONST.BEGIN_ENTITY,
                BB3_CONST.BEGIN_MEMORY,
                BB3_CONST.TOKEN_KNOWLEDGE,
            ]
        )
        context = context.split('\n')
        context = [c for c in context if BB3_CONST.DUMMY_TEXT.lower() not in c.lower()]
        if 'DialogueFrom' not in message['id']:
            context = [
                c
                for c in context
                if not ('your persona: ' in c and BB3_CONST.BEGIN_MEMORY not in c)
            ]
        knowledge = context[-2]
        if BB3_CONST.BEGIN_ENTITY in knowledge:
            begin, end = BB3_CONST.BEGIN_ENTITY, BB3_CONST.END_ENTITY
            new_prefix = PROMPT.CONTEXTUAL_KNOWLEDGE_PREFIX
        elif BB3_CONST.BEGIN_MEMORY in knowledge:
            begin, end = BB3_CONST.BEGIN_MEMORY, BB3_CONST.END_MEMORY
            new_prefix = PROMPT.MEMORY_KNOWLEDGE_PREFIX
        else:
            assert BB3_CONST.TOKEN_KNOWLEDGE in knowledge
            begin, end = BB3_CONST.TOKEN_KNOWLEDGE, BB3_CONST.TOKEN_END_KNOWLEDGE
            new_prefix = PROMPT.SEARCH_KNOWLEDGE_PREFIX
        knowledge = knowledge.replace(begin, f"{new_prefix}:")
        knowledge = knowledge.replace(f" {end}", '')
        knowledge = knowledge.replace(
            'your persona: ', f"{PROMPT.SELF_MEMORY_PREFIX}: "
        )
        knowledge = knowledge.replace(
            'partner\'s persona: ', f"{PROMPT.PARTNER_MEMORY_PREFIX}: "
        )
        message.force_set('text', '\n'.join(context[:-2] + [knowledge] + [context[-1]]))
        return message


@register_mutator("format_vanilla_dialogue_for_decoder_only")
class FormatVanillaDialogueForDecoderOnlyMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        text = message['text'].split('\n')
        new_text = []
        for t in text:
            if 'your persona:' in t:
                t = t.replace('your persona', PROMPT.SELF_MEMORY_PREFIX)
            if t != BB3_CONST.DUMMY_TEXT:
                new_text.append(t)
        message.force_set('text', '\n'.join(new_text))
        return message


@register_mutator("format_light_tasks_for_decoder_only")
class FormatLIGHTForDecoderOnlyMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        text = message['text']
        text = (
            text.replace('_setting_name', f"{PROMPT.LIGHT_SETTING_NAME}:")
            .replace(
                '_setting_desc',
                f"{PROMPT.LIGHT_SETTING_DESC}:",
            )
            .replace(
                '_partner_name',
                f"{PROMPT.LIGHT_PARTNER_NAME}:",
            )
            .replace(
                '_self_name',
                f"{PROMPT.LIGHT_SELF_NAME}:",
            )
            .replace(
                '_self_persona',
                f"{PROMPT.SELF_MEMORY_PREFIX}:",
            )
        )
        message.force_set('text', text)
        return message


@register_mutator("format_style_grounding_tasks_for_decoder_only")
class FormatStyleTasksForDecoderOnlyMutator(MessageMutator):
    def message_mutation(self, message: Message) -> Message:
        text = message['text']
        text = (
            text.replace(BB3_CONST.BEGIN_STYLE, f"{PROMPT.STYLE_PREFIX}:")
            .replace(f" {BB3_CONST.END_STYLE}", '')
            .replace('your persona:', f"{PROMPT.SELF_MEMORY_PREFIX}:")
        )
        message.force_set('text', text)
        return message



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base R2C2 400M Model.

3 Pre-training datasets (pushshift.io Reddit, RoBERTa, & CC)
"""
from parlai.zoo.seeker.seeker_download import download_with_model_type


def download(datapath):
    download_with_model_type(datapath, 'r2c2_base_400M', 'v1.0')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import argparse
import os
from abc import ABC, abstractmethod
from datetime import datetime
import json
from typing import Any, Dict, List, Union

import pandas as pd

from parlai.core.opt import Opt
import parlai.utils.logging as logging

# Defining the class only if Mephisto is installed, since it relies on Mephisto
try:
    from mephisto.abstractions.databases.local_database import LocalMephistoDB
    from mephisto.data_model.unit import Unit
    from mephisto.tools.data_browser import DataBrowser as MephistoDataBrowser
except ImportError:
    pass


class AbstractResultsCompiler(ABC):
    """
    Abstract class for compiling results of crowdsourcing runs.

    Currently only provides utility attributes/methods for analyzing turn annotations.
    """

    @classmethod
    def setup_args(cls):
        parser = argparse.ArgumentParser('Compile crowdsourcing results')
        parser.add_argument(
            '--output-folder', type=str, help='Folder to save output files to'
        )
        parser.add_argument(
            '--results-format',
            type=str,
            choices=['csv', 'json'],
            default='csv',
            help='Output format for results data',
        )
        parser.add_argument(
            '--task-name', type=str, help='Name of the Mephisto task to open'
        )
        parser.add_argument(
            '--database-path',
            type=str,
            default=None,
            help='Path to local Mephisto database. Leave empty for default location.',
        )
        return parser

    def __init__(self, opt: Opt):
        self.task_name = opt['task_name']
        self.output_folder = opt['output_folder']
        self.results_format = opt.get('results_format', 'json')
        self.database_path = opt['database_path']

        # We lazily load these later, or inject their mock version during testing.
        self._mephisto_db = None
        self._mephisto_data_browser = None

    def get_mephisto_data_browser(self) -> MephistoDataBrowser:
        if not self._mephisto_data_browser:
            db = self.get_mephisto_db()
            self._mephisto_data_browser = MephistoDataBrowser(db=db)
        return self._mephisto_data_browser

    def get_mephisto_db(self) -> LocalMephistoDB:
        if not self._mephisto_db:
            self._mephisto_db = LocalMephistoDB(self.database_path)
        return self._mephisto_db

    def get_results_path_base(self) -> str:
        """
        Return the save path for the results file, not including the file extension.
        """
        now = datetime.now()
        return os.path.join(
            self.output_folder,
            f'{self.__class__.__name__}__{now.strftime("%Y%m%d_%H%M%S")}',
        )

    def get_worker_name(self, worker_id: str) -> str:
        """
        Gets the global id of a worker from their Mephisto worker_id.

        The worker_id is the unique id that the crowdsourcing platforms (eg, Amazon
        Mechanical Turk) assign to a single human worker in their system.
        """
        db = self.get_mephisto_db()
        return db.get_worker(worker_id)["worker_name"]

    def get_task_units(self) -> List[Unit]:
        """
        Retrieves the list of work units from the Mephisto task.
        """
        data_browser = self.get_mephisto_data_browser()
        return data_browser.get_units_for_task_name(self.task_name)

    def get_data_from_unit(self, unit: Unit) -> Dict[str, Any]:
        """
        Retrieves task data for a single unit.
        """
        try:
            data_browser = self.get_mephisto_data_browser()
            return data_browser.get_data_from_unit(unit)
        except (IndexError, AssertionError) as error:
            logging.error(error)
            logging.warning(
                f'Skipping unit {unit.db_id}. No message found for this unit.'
            )

    def get_task_data(self) -> List[Dict[str, Any]]:
        """
        Retrieves task data for a list of Mephisto task units.
        """
        task_data = []
        for unit in self.get_task_units():
            unit_data = self.get_data_from_unit(unit)
            if unit_data and self.is_unit_acceptable(unit_data):
                task_data.append(unit_data)

        return task_data

    def is_unit_acceptable(self, unit_data: Dict[str, Any]) -> bool:
        """
        Helps filtering units that are compiled. Override for use.

        Returning False means that the unit data will be discarded.
        """
        if not unit_data:
            # Add your task-specific qualificaiton logic that justifies
            # discarding this unit, based on it data content.
            return False

        return True

    @abstractmethod
    def compile_results(self) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Method for returning the final results as a dataframe or a json.

        For Dict output each key is a unique identifier (eg Assignment ID) for a unit of
        crowdsourcing work. The data for that unit is stored in the value as dictionary.

        Each row of the dataframe consists of one utterance of one conversation, or crowdsourcing interaction.

        NOTE: Preference for new projects is Dict output (see the TODO below).
        TODO: Only support Dict. Deprecate ` pd.DataFrame` when no other code is relying on it.
        """

    def _validate_compiled_result_type(self, results):
        assert isinstance(results, dict) or isinstance(results, pd.DataFrame), (
            'The output of result compiler needs to be a dictionary or a pandas dataframe. '
            f'Found ({type(results)})'
        )

    def compile_and_save_results(self):
        """
        Compile results and save them.

        Results will be saved in the format given by --results-format.
        """
        compiled_results = self.compile_results()
        self._validate_compiled_result_type(compiled_results)
        results_path_base = self.get_results_path_base()
        results_path = f'{results_path_base}.{self.results_format}'
        os.makedirs(self.output_folder, exist_ok=True)
        if self.results_format == 'csv':
            if not isinstance(compiled_results, pd.DataFrame):
                logging.warning(
                    "The requested data output format was 'csv' while the data was compiled as a 'dict'. "
                    'Transforming dictionary data into pd.DataFrame using pandas.'
                )
                compiled_results = pd.DataFrame.from_dict(
                    compiled_results, orient='index'
                )
            compiled_results.to_csv(results_path, index=False)
        elif self.results_format == 'json':
            if isinstance(compiled_results, pd.DataFrame):
                logging.warning(
                    "The requested data output format was 'json' while the data was compiled as a 'dataframe'. "
                    'Transforming dataframe into json using pandas.'
                )
                # Reset the index to make each row have a unique index value
                compiled_results.reset_index().to_json(results_path)
            else:
                with open(results_path, 'w') as fout:
                    fout.write(json.dumps(compiled_results))

        else:
            raise ValueError(
                f'Results save format of "{self.results_format}" currently unsupported!'
            )
        logging.info(f'Wrote results file to {results_path}.')


class AbstractTurnAnnotationResultsCompiler(AbstractResultsCompiler):
    """
    Results compiler subclass to provide utility code for turn annotations.

    Currently incompatible with Mephisto's DataBrowser: all subclasses load results
    files directly from disk.
    TODO: make all subclasses compatible with DataBrowser
    """

    @classmethod
    def setup_args(cls):
        parser = super().setup_args()
        parser.add_argument(
            '--problem-buckets',
            type=str,
            help='Comma-separated list of buckets used for annotation. Set to an empty string to not analyze problem buckets.',
            default='bucket_0,bucket_1,bucket_2,bucket_3,bucket_4,none_all_good',
        )
        return parser

    def __init__(self, opt: Opt):

        super().__init__(opt)

        # Handle inputs
        if opt['problem_buckets'].lower() not in ['', 'none']:
            self.use_problem_buckets = True
            self.problem_buckets = opt['problem_buckets'].split(',')
        else:
            self.use_problem_buckets = False
            self.problem_buckets = []



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import itertools
import torch
from typing import Optional, List, Dict, Any

from parlai.core.agents import Agent, create_agent
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_generator_agent import PPLMetric
import parlai.utils.logging as logging

from projects.safety_bench.utils.wrapper_loading import register_model_wrapper

from projects.bb3.agents.module import Module
import projects.bb3.prompts as PROMPT

from projects.bb3.agents.utils import APIUtils

PROMPTS = {
    "convai2": "A conversation between two persons. Person 2's personality is described.\n\n",
    "wow": "A conversation between two persons. Person 1 is trying to learn about ",
    "none": None,  # no prompt
}


class SimplePromptHistory(object):
    SPEAKER_SELF = "Person 2"
    SPEAKER_OTHER = "Person 1"

    def __init__(self, prompt: Optional[str] = None):
        self.turns = []
        self.prompt = prompt
        self._will_clear = False

    def observe_self(self, text):
        self.turns.append(f'{self.SPEAKER_SELF}: {text}')

    def observe_other(self, text: str):
        assert text is not None
        lines = text.split("\n")
        assert len(lines) >= 1
        while lines and 'your persona' in lines[0]:
            self.turns.append(lines.pop(0))
        end_of_context = len(self.turns)
        speakers = itertools.cycle([self.SPEAKER_OTHER, self.SPEAKER_SELF])
        for speaker, line in zip(speakers, lines[::-1]):
            if '__SILENCE__' in line:
                continue
            self.turns.insert(end_of_context, f'{speaker}: {line}')

    def render_prompt(self) -> str:
        flattened = "\n".join(self.turns + [f'{self.SPEAKER_SELF}:'])
        if self.prompt:
            flattened = f'{self.prompt}{flattened}'
        return flattened

    def get_history_str(self):
        return '\n'.join(self.turns)

    def clear(self):
        self.turns.clear()
        self._will_clear = False

    def prepare_clear(self):
        self._will_clear = True

    def finish_clear(self):
        if self._will_clear:
            self.clear()


class BB3PromptHistory(SimplePromptHistory):
    def __init__(
        self,
        prompt: Optional[str] = None,
        opt: Optional[Opt] = None,
        dictionary: Optional[DictionaryAgent] = None,
    ):
        super().__init__(prompt)
        assert opt is not None
        self.add_speaker_prefixes = opt.get('add_speaker_prefixes', True)
        self.SPEAKER_SELF = (
            opt.get('self_prefix', self.SPEAKER_SELF) or self.SPEAKER_SELF
        )
        self.SPEAKER_OTHER = (
            opt.get('partner_prefix', self.SPEAKER_OTHER) or self.SPEAKER_OTHER
        )
        self.max_prompt_len = opt.get('max_prompt_len', PROMPT.MAX_PROMPT_LEN)
        self.module = Module(opt['module'])
        self.dictionary = dictionary
        self.memories_included = self.module is Module.MEMORY_DECISION and opt.get(
            'memory_decision_use_memories', False
        )
        self.one_turn_history = (
            self.module.is_one_turn_history() and not self.memories_included
        )
        self.prompt = ''
        if opt['include_prompt']:
            if opt['all_vanilla_prompt']:
                self.prompt = Module(Module.VANILLA_DIALOGUE).opt_prompt()
            else:
                self.prompt = self.module.opt_prompt()

        self.shots = self.module.opt_shots()
        if opt.get('num_shots') is not None:
            effective_shots = opt['num_shots'] if opt['num_shots'] >= 0 else 100000
            if effective_shots == 0:
                self.shots = ''
            elif effective_shots > 0:
                self.shots = (
                    '\n\n'.join(self.shots.split('\n\n')[:effective_shots]) + '\n\n'
                )
        self.final_prefix = f"{self.module.opt_final_prefix()}:"
        if opt.get('final_prefix'):
            self.final_prefix = f"{opt['final_prefix']}:"
        if opt.get('final_prefix_space'):
            self.final_prefix = f"{self.final_prefix} "
        self.pre_context_tok = self.module.opt_pre_context_tok()
        self.post_context_tok = self.module.opt_post_context_tok()
        self.style_string = (
            f"{PROMPT.STYLE_PREFIX}: {opt['insert_style']}"
            if opt.get('insert_style')
            else ''
        )
        self.debug = opt.get('debug_bb3', False)
        if not opt.get('include_substitution_tokens', True):
            self.pre_context_tok = ''
            self.post_context_tok = ''

    def observe_self(self, text):
        if self.add_speaker_prefixes:
            super().observe_self(text)
            return

        self.turns.append(text)

    def observe_other(self, text: str):
        assert text is not None
        lines = text.split("\n")
        assert len(lines) >= 1
        while lines and any(
            p in lines[0]
            for p in [
                'your persona',
                f"{self.SPEAKER_SELF}'s Persona",
                f"{self.SPEAKER_OTHER}'s Persona",
            ]
        ):
            self.turns.append(lines.pop(0))
        end_of_context = len(self.turns)
        speakers = itertools.cycle([self.SPEAKER_OTHER, self.SPEAKER_SELF])
        for speaker, line in zip(speakers, lines[::-1]):
            if (
                '__SILENCE__' in line
                or line == PROMPT.OPENING_PREFIX
                or not line.strip()
            ):
                continue
            insert_line = f"{speaker}: {line}" if self.add_speaker_prefixes else line
            self.turns.insert(end_of_context, insert_line)

    def render_flattened(self, turns: List[str]) -> str:
        """
        A prompt consists of the following components:

        - Prompt: the instruction to the model, e.g., "A conversation between two persons"
        - Shots: the k-shot examples to show the model
        - Pre-context token: this token is replaced at runtime with appropriate pre-context text; e.g., for the search knowledge model, this is replaced with external knowledge from the internet
        - Turns: these are the dialogue turns seen so far
        - Post-context token: this token is replaced at runtime with an appropriate post-context text; e.g., for the search dialogue model, this is replaced with the generated knowledge from the search knowledge model
        - Final prefix: the token after which the model will generate a response. This varies between module.
        """
        if not turns:
            # opening dialogue; just render the prefix
            return self.final_prefix
        if turns[-1].strip() == self.final_prefix.strip():
            # some tasks put the final prefix in for us; we don't want that!
            turns = turns[:-1]
        if self.one_turn_history:
            turns = turns[-1:]
        elif self.memories_included and len(turns) > 1:
            memory_turns = [
                t for t in turns if t.startswith(PROMPT.MEMORY_KNOWLEDGE_PREFIX)
            ]
            turns = memory_turns + turns[-1:]
        flattened_turns = "\n".join(turns)
        post_context = f"\n{self.post_context_tok}" if self.post_context_tok else ''
        pre_context = f"{self.pre_context_tok}\n" if self.pre_context_tok else ''
        style = f"\n{self.style_string}" if self.style_string else ''
        shots = self.shots if self.shots else ''
        final = f'{self.prompt}{shots}{pre_context}{flattened_turns}{post_context}{style}\n{self.final_prefix}'
        return final

    def render_prompt(self) -> str:
        assert self.dictionary is not None
        turn_idx = 0
        flattened = self.render_flattened(self.turns)
        while len(self.dictionary.txt2vec(flattened)) >= self.max_prompt_len:
            turn_idx += 1
            flattened = self.render_flattened(self.turns[turn_idx:])
        if self.debug:
            logging.info(f'Module: {self.module}')
            logging.info(flattened)
        return flattened


class SimpleOPTAgent(Agent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser.add_argument(
            '--prompt',
            choices=PROMPTS.keys(),
            default='none',
            help='Pre-made prompts. Use --raw-prompt to manually write one.',
        )
        parser.add_argument(
            '--self-prefix',
            default=PROMPT.SELF_PREFIX,
            help="Prefix token indicating 'self' turn",
        )
        parser.add_argument(
            '--partner-prefix',
            default=PROMPT.PARTNER_PREFIX,
            help="Prefix token indicating 'partner' turn",
        )
        parser.add_argument(
            '--final-prefix', default=None, help="Specify to override the final prefix."
        )
        parser.add_argument(
            '--final-prefix-space',
            default=False,
            type='bool',
            help="Specify to include a space after the final prefix.",
        )
        parser.add_argument(
            '--raw-prompt', default=None, help='Use to manually specify a raw prompt.'
        )
        parser.add_argument(
            '--inference',
            default='greedy',
            choices=[
                'nucleus',
                'beam',
                'greedy',
                'sample_and_rank',
                'factual_nucleus',
                'sample_and_rank_factual_nucleus',
            ],
        )
        parser.add_argument(
            '--penalize-repetitions',
            default=False,
            type='bool',
            help="""
                Penalize repetitions according to heuristics put forth by OpenAI
                From https://beta.openai.com/docs/api-reference/engines/retrieve:

                mu[j] -> mu[j] - c[j] * alpha_frequency - float(c[j] > 0) * alpha_presence
                Where:

                mu[j] is the logits of the j-th token
                c[j] is how often that token was sampled prior to the current position
                float(c[j] > 0) is 1 if c[j] > 0 and 0 otherwise
                alpha_frequency is the frequency penalty coefficient
                alpha_presence is the presence penalty coefficient
            """,
        )
        parser.add_argument(
            '--penalize-ctxt-repetitions',
            default=False,
            type='bool',
            help='apply repetition blocking heuristics to context as well.',
        )
        parser.add_argument('--skip-generation', default=False, type='bool')
        parser.add_argument('--beam-size', default=1, type=int)
        parser.add_argument('--beam-min-length', default=0, type=int)
        parser.add_argument('--beam-max-length', default=32, type=int)
        parser.add_argument('--topp', default=0.9, type=float)
        parser.add_argument(
            '--lambda-decay',
            default=0.9,
            type=float,
            help='hyperparameter in factual_nucleus inference; decay p value at this rate',
        )
        parser.add_argument(
            '--omega-bound',
            default=0.3,
            type=float,
            help='lower bound of p value in factual nucleus, after decay',
        )
        parser.add_argument(
            '--alpha-presence',
            default=0.5,
            type=float,
            help='penalty applied to the logits for presence of previous tokens in the generation',
        )
        parser.add_argument(
            '--alpha-frequency',
            default=0.5,
            type=float,
            help='penalty applied to the logits for frequency of previous tokens in the generation',
        )
        parser.add_argument(
            '--alpha-presence-src',
            default=0.5,
            type=float,
            help='penalty applied to the logits for presence of previous tokens in the context',
        )
        parser.add_argument(
            '--alpha-frequency-src',
            default=0.5,
            type=float,
            help='penalty applied to the logits for frequency of previous tokens in the context',
        )
        parser.add_argument('--temperature', default=1.0, type=float)
        parser.add_argument('--stop-token', default="\n", type=str)
        parser.add_argument('--server', default=APIUtils.DEFAULT_SERVER, type=str)
        parser.add_argument(
            '--max-retry-api',
            default=-1,
            type=int,
            help='Number of times to retry on API request failures (< 0 for unlimited retry).',
        )
        parser.add_argument(
            '--server-timeout',
            default=APIUtils.DEFAULT_SERVER_TIMEOUT,
            type=int,
            help='Timeout (s) for the API request to the GPTZ/OPT workers.',
        )
        return parser

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared=shared)
        self.id = "opt"
        self.server = opt['server']
        self.n_retry_api_exception = opt['max_retry_api']
        self.server_timeout = opt['server_timeout']
        self.api_key = opt.get('api_key', APIUtils.DEFAULT_KEY)
        if opt.get('raw_prompt'):
            prompt = opt['raw_prompt']
        elif opt['prompt']:
            prompt = PROMPTS[opt['prompt']]
        else:
            prompt = None
        if shared is None:
            logging.debug(f"GPT-Z setting prompt to '{prompt}'")
        self.passed_in_prompt = prompt
        self.history = SimplePromptHistory(prompt=prompt)
        self.request_delay = opt.get('request_delay', 0.5)

    def observe(self, obs):
        if not obs.get('batch_padding'):
            self.history.observe_other(obs.get('text', ''))

        if obs.get('episode_done'):
            self.history.prepare_clear()

        obs['prompt'] = self.history.render_prompt()

        return super().observe(obs)

    def reset(self):
        super().reset()
        self.history.clear()

    def act(self):
        response = self.batch_act([self.observation])[0]
        self.self_observe(response)
        return response

    def build_prompt_label(self, obs: Message, label: str) -> str:
        """
        Builds the prompt label concatenation.

        :param obs:
            observation we're dealing with
        :param label:
            label for example

        :return prompt_label:
            return the prompt_label concatenation
        """
        return obs['prompt'] + " " + label

    def get_echo_results(self, observations: List[Message]):
        """
        Get echo results from the server.
        """
        results = APIUtils.request_many(
            server=self.server,
            api_key=self.api_key,
            prompts=[o['prompt_label'] for o in observations],
            timeout=self.server_timeout,
            max_num_tries=self.n_retry_api_exception,
            max_tokens=0,
            echo=True,
            request_delay=self.request_delay,
        )
        return results

    def get_gen_results(self, observations: List[Message], **gen_params):
        """
        Get GEN results from the server.
        """
        results = APIUtils.request_many(
            server=self.server,
            api_key=self.api_key,
            prompts=[o['prompt'] for o in observations],
            timeout=self.server_timeout,
            max_num_tries=self.n_retry_api_exception,
            request_delay=self.request_delay,
            **gen_params,
        )
        return results

    def rank_samples(self, result: Dict[str, Any]) -> str:
        """
        Rank nucleus sampled outputs according to perplexities.

        Inference: sample_and_rank
            Sampled generations are ranked according to their returned token probabilities.
            For this, we use the probabilities returned by metaseq.

        :param result:
            a single result from API call.
        """
        texts_i = [r['text'] for r in result['choices']]
        ppls = []
        for choice in result['choices']:
            logprobs = choice['logprobs']['token_logprobs']
            if logprobs[0] > 0:
                logprobs = logprobs[1:]
            ppls.append(PPLMetric(-sum(logprobs), len(logprobs)).value())
        ppls = torch.tensor(ppls)
        return texts_i[ppls.argmin()]

    def batch_act(self, observations):
        original_observations = observations

        observations = [o for o in observations if 'batch_padding' not in o]
        messages = [Message(id=self.id) for o in observations]

        # possible ppl evaluations
        for o in observations:
            if 'labels' in o:
                label = o['labels'][0]
            elif 'eval_labels' in o:
                label = o['eval_labels'][0]
            else:
                break
            complete = self.build_prompt_label(o, label)
            o.force_set('prompt_label', complete)
        else:
            # for-else means we execute this unless we were missing labels
            results = self.get_echo_results(observations)
            label_ppls, all_ppls = APIUtils.compute_perplexities(observations, results)
            for ppl, ctxt_ppl, msg in zip(label_ppls, all_ppls, messages):
                msg['metrics'] = {'ppl': ppl, 'ctxt_label_ppl': ctxt_ppl}

        # actual generations
        if not self.opt['skip_generation']:
            if self.opt['inference'] == 'greedy':
                self.opt['beam_size'] = 1
                self.opt['topp'] = -1.0
            elif self.opt['inference'] == 'beam':
                self.opt['topp'] = -1.0
            if 'factual_nucleus' not in self.opt['inference']:
                self.opt['lambda_decay'] = -1
            if not self.opt['penalize_repetitions']:
                self.opt['alpha_presence'] = 0
                self.opt['alpha_frequency'] = 0
            if not self.opt['penalize_ctxt_repetitions']:
                self.opt['alpha_presence_src'] = 0
                self.opt['alpha_frequency_src'] = 0

            gen_params = {
                'best_of': self.opt['beam_size'],
                'top_p': self.opt['topp'],
                'temperature': self.opt['temperature'],
                'min_tokens': self.opt['beam_min_length'],
                'max_tokens': self.opt['beam_max_length'],
                'stop': self.opt['stop_token'],
                'echo': False,
                'lambda_decay': self.opt['lambda_decay'],
                'omega_bound': self.opt['omega_bound'],
                'alpha_presence': self.opt['alpha_presence'],
                'alpha_frequency': self.opt['alpha_frequency'],
                'alpha_presence_src': self.opt['alpha_presence_src'],
                'alpha_frequency_src': self.opt['alpha_frequency_src'],
            }
            logging.debug([o['prompt'] for o in observations])

            results = self.get_gen_results(observations, **gen_params)
            # post process out if the model generated multiple lines
            for m, r in zip(messages, results):
                if APIUtils.is_request_failed_response(r):
                    m['text'] = APIUtils.METASEQ_FAIL_MESSAGE_TEXT
                    m['failures'] = r['failures']
                else:
                    if self.opt['inference'] in (
                        'sample_and_rank',
                        'sample_and_rank_factual_nucleus',
                    ):
                        m['text'] = self.rank_samples(r)
                    else:
                        m['text'] = r['choices'][0]['text'].strip()
                    m['logprobs'] = sum(r['choices'][0]['logprobs']['token_logprobs'])

        # we might have batch padding that we skipped earlier. collate it back.
        messages_out = []
        for o in original_observations:
            if 'batch_padding' in o:
                messages_out.append(Message())
            else:
                messages_out.append(messages.pop(0))
        return messages_out

    def self_observe(self, self_message):
        if 'labels' in self.observation:
            label = self.observation['labels'][0]
        elif 'eval_labels' in self.observation:
            label = self.observation['eval_labels'][0]
        elif 'text' in self_message:
            label = self_message['text']
        else:
            return
        self.history.observe_self(label)
        self.history.finish_clear()


class BB3OPTAgent(SimpleOPTAgent):
    """
    Override several components for fine-grained control.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = SimpleOPTAgent.add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            '--module', type=str, choices=[m.tag() for m in Module], help='which module'
        )
        parser.add_argument(
            '--include-substitution-tokens',
            type='bool',
            default=True,
            help='specify false to remove the pre-context/post-context substitution tokens. '
            'useful for evaluating tasks directly with this agent.',
        )
        parser.add_argument(
            '--include-prompt',
            type='bool',
            default=True,
            help='Whether to include prompt',
        )
        parser.add_argument(
            '--num-shots',
            type=int,
            default=-1,
            help='how many k-shot examples to put in the prompt. Default <0 is all',
        )
        parser.add_argument(
            '--max-prompt-len',
            type=int,
            default=PROMPT.MAX_PROMPT_LEN,
            help='Longest sequence to send to API',
        )
        parser.add_argument(
            '--all-vanilla-prompt',
            type='bool',
            default=False,
            help='If True, and --include-prompt, then all prompts are just the vanilla, "a conversation", prompts.',
        )
        parser.add_argument(
            '--insert-style',
            type=str,
            default=None,
            help='if set, will insert this style before the final response',
        )
        parser.add_argument(
            '--add-speaker-prefixes',
            type='bool',
            default=True,
            help='Whether to prefix utterances with speaker tokens',
        )
        parser.add_argument(
            '--generation-take-last-newline',
            type='bool',
            default=True,
            help='if a generation is returned with a newline character, set True to take last newline. '
            'set False to take first new line.',
        )
        parser.add_argument(
            '--generation-allow-newline',
            type='bool',
            default=False,
            help='if a generation is returned with a newline character, set True to return all generated tokens.',
        )
        parser.add_argument(
            '--memory-decision-use-memories',
            type='bool',
            default=False,
            help='If true, the memory decision module will have access to memories when making the decision',
        )
        parser.add_argument(
            '--exclude-knowledge-from-ctxt-penalty',
            type='bool',
            default=False,
            help='Specified in tandem with --alpha-presence/frequency-src. '
            'If true, we block from context, but *not* including anything after knowledge.',
        )
        parser.add_argument('--debug-bb3', type='bool', default=False, dest='debug_bb3')
        return parser

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.max_prompt_len = opt.get('max_prompt_len', PROMPT.MAX_PROMPT_LEN)
        if not shared:
            self.dictionary = DictionaryAgent(
                {**ParlaiParser(True, True).parse_args([]), 'dict_tokenizer': 'gpt2'}
            )
        else:
            self.dictionary = shared['dictionary']
        self.history = BB3PromptHistory(opt=opt, dictionary=self.dictionary)
        self.module = Module(opt['module'])

    def share(self):
        shared = super().share()
        shared['dictionary'] = self.dictionary
        return shared

    def set_retriever_type(self, r_type: str):
        """
        No-op.
        """
        pass

    def build_prompt_label(self, obs: Message, label: str) -> str:
        """
        Override to ensure we don't run up against the 2048 truncation.

        :param obs:
            observation we're dealing with
        :param label:
            label for example

        :return prompt_label:
            return the prompt_label concatenation, with truncation minded.
        """
        complete = obs['prompt'] + " " + label
        while len(self.dictionary.txt2vec(complete)) > self.max_prompt_len:
            obs.force_set('prompt', '\n'.join(obs['prompt'].split('\n')[1:]))
            complete = obs['prompt'] + " " + label
        return complete

    def _get_alpha_src_penalty_end_idx(self, observations: List[Message]) -> int:
        """
        Determine where we stop the penalty on context blocking.

        :param observations:
            observations with prompts

        :return end_idx:
            returns end idx.
        """
        if (
            len(observations) > 1
            or not self.opt['exclude_knowledge_from_ctxt_penalty']
            or not self.module.is_dialogue()
        ):
            return -1

        obs = observations[0]
        all_tokens = self.dictionary.txt2vec(obs['prompt'])
        sub_tokens = self.dictionary.txt2vec(
            self.module.opt_dialogue_knowledge_prefix().strip()
        )
        end_idx = -1
        for i in list(range(len(all_tokens) - len(sub_tokens)))[::-1]:
            # iterate backwards since this should be quicker.
            if i < len(sub_tokens):
                break
            if all_tokens[i - len(sub_tokens) : i] == sub_tokens:
                end_idx = i - len(sub_tokens)
                break

        return end_idx

    def get_echo_results(self, observations: List[Message]):
        """
        Override to ensure that we get the results we need, since sometimes things are
        dropped.
        """
        results = super().get_echo_results(observations)
        while any([result['choices'][0]['text'] == '' for result in results]):
            results = super().get_echo_results(observations)
        return results

    def get_gen_results(self, observations: List[Message], **gen_params):
        """
        Cleans up the generated text from the OPT/GPTz agent.
        """
        results = super().get_gen_results(
            observations,
            **gen_params,
            alpha_src_penalty_end_idx=self._get_alpha_src_penalty_end_idx(observations),
        )

        for r in results:
            if not APIUtils.is_request_failed_response(r):
                r['choices'][0]['text'] = r['choices'][0]['text'].strip("\n")

        if not self.opt.get('generation_allow_newline') and any(
            '\n' in res['choices'][0]['text']
            for res in results
            if not APIUtils.is_request_failed_response(res)
        ):
            if self.opt.get('generation_take_last_newline', True):
                logging.warning("Generation contains newline; taking last utterance")
            else:
                logging.warning("Generation contains newline; taking first utterance")
            for result in results:
                if APIUtils.is_request_failed_response(result):
                    continue
                result['choices'][0]['text'] = (
                    result['choices'][0]['text'].split('\n')[-1].split(":")[-1]
                    if self.opt.get('generation_take_last_newline', True)
                    else result['choices'][0]['text'].split('\n')[0].split(":")[-1]
                )
        return results

    def batch_act(self, observations):
        """
        Adapted to the sequential calls in BB and the potential failure on calls.

        Only sends healthy observations to the parent class for inference. Returns the
        unhealthy (failed) observatios without change.
        """
        failed_observations_indx = []
        needs_inference = []

        # Filtering out the observations that have failures in their previous inference rounds.
        for ix, obs in enumerate(observations):
            if APIUtils.is_request_failed_response(obs):
                failed_observations_indx.append(ix)
            else:
                # Only sending those that are still healthy in their flow to the main model for inference.
                needs_inference.append(obs)
        model_resp = super().batch_act(needs_inference)
        batch_act_response = []

        act_observation_idx = 0
        for idx, obs in enumerate(observations):
            if idx in failed_observations_indx:
                batch_act_response.append(observations[idx])
            else:
                # Making sure that the generated response is aligne with its corresponding observations.
                assert (
                    obs == needs_inference[act_observation_idx]
                ), 'Responeses generated on valid observations of model do not align'
                batch_act_response.append(model_resp[act_observation_idx])
                act_observation_idx += 1
        return batch_act_response


class MockOptAgent(BB3OPTAgent):

    mock_text: str
    mock_score: List[int]

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.mock_text = 'I am a chatbot'
        self.mock_score = [-1] * 4
        if self.history.module is Module.SEARCH_DECISION:
            self.mock_text = 'search'
            self.mock_score = [-1]
        elif self.history.module is Module.MEMORY_DECISION:
            self.mock_text = 'access memory'
            self.mock_score = [-1] * 2

    def get_echo_results(self, observations: List[Message]):
        return [] * len(observations)

    def get_gen_results(self, observations: List[Message], **gen_params):
        return [
            {
                'choices': [
                    {
                        'text': self.mock_text,
                        'logprobs': {'token_logprobs': self.mock_score},
                    }
                ]
            }
            for _ in range(len(observations))
        ]


@register_model_wrapper("opt")
class OPTSafetyWrapper(object):
    prompt = None

    def __init__(self):
        self.agent = create_agent(
            {
                'model': 'projects.bb3.agents.opt_api_agent:SimpleOPTAgent',
                'prompt': self.prompt,
            }
        )

    def get_response(self, text: str) -> str:
        return self.agent.respond(text)


@register_model_wrapper("opt_convai2")
class OPTConvai2SafetyWrapper(OPTSafetyWrapper):
    prompt = 'convai2'


@register_model_wrapper("opt_wow")
class OPTWowSafetyWrapper(OPTSafetyWrapper):
    prompt = 'wow'


class HistoryRaw(SimplePromptHistory):
    """
    Version of history that does *not* include the assumptions of dialogue.
    """

    def observe_self(self, text):
        self.turns.append(text)

    def observe_other(self, text):
        self.turns.append(text)

    def render_prompt(self):
        res = "".join(self.turns)
        if self.prompt:
            res = f'{self.prompt}{res}'
        return res

    def get_history_str(self):
        return ''.join(self.turns)


class OPTRawAgent(SimpleOPTAgent):
    """
    Version of GPTZ agent that does *not* include the assumptions of dialogue.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.history = HistoryRaw(prompt=self.passed_in_prompt)
